"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from contextlib import suppress
from re import compile
from typing import Literal
from typing import TYPE_CHECKING

from encommon.utils import rgxp_match

from .models import BltnModels
from ..homie.plugins import HomieDriver
from ..utils import InvalidParam

if TYPE_CHECKING:
    from .params import DriverBltnRegexpParams
    from ..homie.plugins import HomiePluginKinds
    from ..homie.threads import HomieStreamItem



class DriverBltnRegexp(HomieDriver):
    """
    Match specific conditions for determining desired state.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        params = self.params

        _rgxps = params.patterns


        for pattern in _rgxps:

            with suppress(Exception):
                compile(pattern)
                continue

            raise InvalidParam(
                param='patterns',
                value=pattern,
                error='invalid')


    @property
    def family(
        self,
    ) -> Literal['builtins']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'builtins'


    @property
    def kinds(
        self,
    ) -> list['HomiePluginKinds']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return ['occur']


    @property
    def params(
        self,
    ) -> 'DriverBltnRegexpParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            BltnModels
            .drivers()
            .regexp())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def occur(
        self,
        sitem: 'HomieStreamItem',
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        :param sitem: Item containing information for operation.
        :returns: Boolean indicating the conditional outcomes.
        """

        params = self.params
        patterns = params.patterns
        complete = params.complete

        return rgxp_match(
            values=str(sitem.event),
            patterns=patterns,
            complete=complete)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Optional

from pydantic import Field

from ...plugins.change import PhueChangeSensor
from ....homie.params.common import HomieParamsModel



class DriverPhueChangeParams(HomieParamsModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    devices: Annotated[
        list[str],
        Field(...,
              description='Devices that are in scope',
              min_length=1)]

    sensors: Annotated[
        Optional[list[PhueChangeSensor]],
        Field(None,
              description='Sensors that will be matched',
              min_length=1)]

    since: Annotated[
        int,
        Field(0,
              description='Minimum time since changed',
              ge=0)]


    def __init__(
        # NOCVR
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        devices = data.get('devices')
        sensors = data.get('sensors')

        if isinstance(devices, str):
            data['devices'] = [devices]

        if isinstance(sensors, str):
            data['sensors'] = [sensors]

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Time
from encommon.times import TimerParams
from encommon.times import Timers
from encommon.times import TimersParams
from encommon.times.params import _TIMERS

from .child import HomieChild
from .helpers import whered
from ..models import HomieModels
from ..plugins import HomieWhere

if TYPE_CHECKING:
    from .device import HomieDevice
    from .group import HomieGroup
    from ..params import HomieDesireParams
    from ..threads import HomieActionNode



class HomieDesire(HomieChild):
    """
    Match specific conditions for determining desired state.
    """

    __wheres: list[HomieWhere]

    __timers: Timers


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__wheres = []

        self.__build_wheres()
        self.__build_timers()


    def __build_timers(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        homie = self.homie
        params = self.params

        timers: _TIMERS = {}

        pause = params.pause
        delay = params.delay

        store = (
            homie.params
            .database)

        _model = TimerParams
        model = TimersParams


        start = (
            'now'
            if delay is True
            else 'min')


        names = params.devices

        for name in names or []:

            unique = f'device/{name}'

            timer = _model(
                timer=pause,
                start=start)

            timers[unique] = timer


        names = params.groups

        for name in names or []:

            unique = f'group/{name}'

            timer = _model(
                timer=pause,
                start=start)

            timers[unique] = timer


        name = self.name
        group = f'desire/{name}'

        object = Timers(
            model(timers),
            store=store,
            group=group)

        self.__timers = object



    def __build_wheres(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params
        wheres = params.wheres
        homie = self.homie

        if wheres is None:
            return None

        model = HomieWhere


        plugins: list[HomieWhere] = []


        for where in wheres:

            object = model(
                homie, where)

            plugins.append(object)


        self.__wheres = plugins


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        wheres = self.__wheres

        for where in wheres:
            where.validate()


    @property
    def kind(
        self,
    ) -> Literal['desire']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'desire'


    @property
    def params(
        self,
    ) -> 'HomieDesireParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            HomieModels
            .desire())

        params = super().params

        assert isinstance(
            params, model)

        return params


    @property
    def weight(
        self,
    ) -> int:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.params.weight


    @property
    def devices(
        self,
    ) -> list['HomieDevice']:
        """
        Return the value for the attribute from class instance.

        .. note::
           This method is identical to same within HomieAspire.

        :returns: Value for the attribute from class instance.
        """

        params = self.params
        names = params.devices

        devices = (
            self.homie.childs
            .devices)

        childs: set['HomieDevice'] = set()

        if names is None:
            return list(childs)

        for name in names:

            child = devices[name]

            childs.add(child)

        return list(childs)


    @property
    def groups(
        self,
    ) -> list['HomieGroup']:
        """
        Return the value for the attribute from class instance.

        .. note::
           This method is identical to same within HomieAspire.

        :returns: Value for the attribute from class instance.
        """

        params = self.params
        names = params.groups

        groups = (
            self.homie.childs
            .groups)

        childs: set['HomieGroup'] = set()

        if names is None:
            return list(childs)

        for name in names:

            child = groups[name]

            childs.add(child)

        return list(childs)


    @property
    def wheres(
        self,
    ) -> list[HomieWhere]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return list(self.__wheres)


    def when(
        self,
        time: Time,
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        .. note::
           Somewhat similar to same method within HomieAspire.

        :param time: Time that will be used in the conditionals.
        :returns: Boolean indicating the conditional outcomes.
        """

        homie = self.homie

        try:

            wheres = (
                whered(self, time))

            if len(wheres) >= 1:
                return all(wheres)

        except Exception as reason:

            homie.logger.log_e(
                base=self,
                name=self,
                status='exception',
                exc_info=reason)

        return False


    def paused(
        self,
        target: 'HomieActionNode',
    ) -> bool:
        """
        Return the boolean indicating whether operation delayed.

        .. note::
           This method is identical to same within HomieAspire.

        :param target: Device or group that in scope for match.
        :returns: Boolean indicating whether operation delayed.
        """

        timers = self.__timers

        name = target.name
        kind = target.kind

        ready = timers.ready(
            f'{kind}/{name}',
            update=False)

        return ready is False


    def matched(
        self,
        target: 'HomieActionNode',
    ) -> None:
        """
        Perform operations when this desire is the most desired.

        .. note::
           Somewhat similar to same method within HomieAspire.

        :param target: Device or group that in scope for match.
        """

        homie = self.homie
        timers = self.__timers

        if homie.dryrun:
            return None

        name = target.name
        kind = target.kind

        timers.update(
            f'{kind}/{name}',
            value='now')


    def omitted(
        self,
        target: Optional['HomieActionNode'] = None,
    ) -> None:
        """
        Perform operations when this desire is the most desired.

        .. note::
           Somewhat similar to same method within HomieAspire.

        :param target: Device or group that in scope for match.
        """

        homie = self.homie
        timers = self.__timers
        params = self.params
        delay = params.delay

        if homie.dryrun:
            return None

        update = (
            'now'
            if delay is True
            else 'min')


        if target is None:

            _timers = (
                timers.children
                .values())

            for timer in _timers:
                timer.update(update)


        elif target is not None:

            name = target.name
            kind = target.kind

            timers.update(
                f'{kind}/{name}',
                value=update)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING
from typing import Type

if TYPE_CHECKING:
    from .action import PhueActionItem
    from .params import DriverPhueButtonParams
    from .params import DriverPhueChangeParams
    from .params import DriverPhueContactParams
    from .params import DriverPhueMotionParams
    from .params import DriverPhueSceneParams
    from .params import PhueOriginParams
    from .update import PhueUpdateItem
    from .stream import PhueStreamItem



class PhueDriverModels:
    """
    Return the class object that was imported within method.
    """


    @classmethod
    def button(
        cls,
    ) -> Type['DriverPhueButtonParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            DriverPhueButtonParams)

        return DriverPhueButtonParams


    @classmethod
    def change(
        cls,
    ) -> Type['DriverPhueChangeParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            DriverPhueChangeParams)

        return DriverPhueChangeParams


    @classmethod
    def contact(
        cls,
    ) -> Type['DriverPhueContactParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            DriverPhueContactParams)

        return DriverPhueContactParams


    @classmethod
    def motion(
        cls,
    ) -> Type['DriverPhueMotionParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            DriverPhueMotionParams)

        return DriverPhueMotionParams


    @classmethod
    def scene(
        cls,
    ) -> Type['DriverPhueSceneParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            DriverPhueSceneParams)

        return DriverPhueSceneParams



class PhueModels:
    """
    Return the class object that was imported within method.
    """


    @classmethod
    def origin(
        cls,
    ) -> Type['PhueOriginParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            PhueOriginParams)

        return PhueOriginParams


    @classmethod
    def action(
        cls,
    ) -> Type['PhueActionItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .action import (
            PhueActionItem)

        return PhueActionItem


    @classmethod
    def update(
        cls,
    ) -> Type['PhueUpdateItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .update import (
            PhueUpdateItem)

        return PhueUpdateItem


    @classmethod
    def stream(
        cls,
    ) -> Type['PhueStreamItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .stream import (
            PhueStreamItem)

        return PhueStreamItem


    @classmethod
    def drivers(
        cls,
    ) -> Type['PhueDriverModels']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        return PhueDriverModels



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING

from encommon.types import DictStrAny

from .thread import HomieThread
from .thread import HomieThreadItem

if TYPE_CHECKING:
    from ..childs import HomieOrigin



@dataclass
class HomieStreamItem(HomieThreadItem):
    """
    Contain information for sharing using the Python queue.
    """

    event: DictStrAny


    def __init__(
        self,
        origin: 'HomieOrigin',
        event: DictStrAny,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        event = deepcopy(event)

        self.event = event

        super().__init__(origin)



class HomieStream(HomieThread):
    """
    Common methods and routines for Homie Automate threads.
    """


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service threads.
        """

        member = self.member
        vacate = member.vacate

        if vacate.is_set():
            return None

        self.operate_streams()


    def operate_streams(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service threads.
        """

        raise NotImplementedError



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING
from typing import get_args

from encommon.types import NCFalse
from encommon.types import getate

from .helpers import phue_sensors
from ..models import PhueModels
from ..stream import PhueStreamItem
from ...homie.plugins import HomieDriver
from ...utils import InvalidParam

if TYPE_CHECKING:
    from ..params import DriverPhueMotionParams
    from ...homie.plugins import HomiePluginKinds
    from ...homie.threads import HomieStreamItem



PhueMotionState = Literal[
    'motion',
    'no_motion']

_STATES = list(
    get_args(PhueMotionState))



class DriverPhueMotion(HomieDriver):
    """
    Match specific conditions for determining desired state.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        plugin = self.plugin
        family = self.family
        homie = plugin.homie
        childs = homie.childs
        devices = childs.devices
        params = self.params

        _device = params.device


        if _device not in devices:

            raise InvalidParam(
                param='device',
                value=_device,
                error='noexist')


        device = devices[_device]

        _family = device.family

        if _family != family:

            raise InvalidParam(
                param='device',
                value=_device,
                error='invalid')


    @property
    def family(
        self,
    ) -> Literal['philips']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'philips'


    @property
    def kinds(
        self,
    ) -> list['HomiePluginKinds']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return ['occur']


    @property
    def params(
        self,
    ) -> 'DriverPhueMotionParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            PhueModels
            .drivers()
            .motion())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def occur(  # noqa: CFQ004
        self,
        sitem: 'HomieStreamItem',
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        :param sitem: Item containing information for operation.
        :returns: Boolean indicating the conditional outcomes.
        """

        plugin = self.plugin
        homie = plugin.homie
        childs = homie.childs
        devices = childs.devices

        params = self.params
        target = params.device
        states = (
            params.states
            or _STATES)


        model = PhueStreamItem

        if not isinstance(sitem, model):
            return NCFalse


        event = sitem.event

        if 'motion' not in event:
            return False


        source = (
            devices[target]
            .source)

        if source is None:
            return False

        service = event['id']


        sensors = (
            phue_sensors(source))

        if 'motion' not in sensors:
            return NCFalse

        intend = sensors['motion']

        if service != intend:
            return False


        valid = getate(
            event,
            'motion/motion_valid')

        if valid is False:
            return NCFalse


        _state = getate(
            event,
            'motion/motion')

        if _state is not None:

            state = 'no_motion'

            if _state is True:
                state = 'motion'

            if state not in states:
                return False


        return _state is not None



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from encommon.colors import Color

from .thread import HomieThread
from .thread import HomieThreadItem

if TYPE_CHECKING:
    from ..childs import HomieDevice
    from ..childs import HomieGroup
    from ..childs import HomieOrigin
    from ..childs import HomieScene
    from ..common import HomieState



HomieActionBase = dict[
    'HomieActionNode',
    'HomieActionItem']

HomieActionNode = Union[
    'HomieGroup', 'HomieDevice']



@dataclass
class HomieActionItem(HomieThreadItem):
    """
    Contain information for sharing using the Python queue.
    """

    group: Optional[str] = None
    device: Optional[str] = None

    state: Optional['HomieState'] = None
    color: Optional[Color] = None
    level: Optional[int] = None
    scene: Optional[str] = None


    def __init__(  # noqa: CFQ002
        self,
        origin: 'HomieOrigin',
        target: HomieActionNode,
        *,
        state: Optional['HomieState'] = None,
        color: Optional[str | Color] = None,
        level: Optional[int] = None,
        scene: Optional['HomieScene'] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        if isinstance(color, str):
            color = Color(color)

        name = target.name
        kind = target.kind

        if kind == 'group':
            self.group = name

        if kind == 'device':
            self.device = name

        self.state = state
        self.color = color
        self.level = level
        self.scene = (
            scene.name
            if scene is not None
            else None)

        super().__init__(origin)



class HomieAction(HomieThread):
    """
    Common methods and routines for Homie Automate threads.
    """


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service threads.
        """

        # Where actions executed

        homie = self.homie
        childs = homie.childs
        origins = childs.origins
        member = self.member
        vacate = member.vacate
        aqueue = self.aqueue

        origin = self.origin
        name = origin.name

        while not aqueue.empty:

            aitem = aqueue.get()

            if vacate.is_set():
                continue

            if self.expired(aitem):
                continue

            _origin = origins[
                aitem.origin]

            _name = _origin.name
            assert name == _name

            self.execute(aitem)


    def execute(
        self,
        aitem: HomieActionItem,
    ) -> None:
        """
        Perform the operation related to Homie service threads.

        :param aitem: Item containing information for operation.
        """

        raise NotImplementedError



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.times import Time

from ..params.common import HomieParamsModel

if TYPE_CHECKING:
    from .common import HomiePluginKinds
    from .plugin import HomiePlugin
    from ..common import HomieFamily
    from ..threads import HomieStreamItem



class HomieDriver:
    """
    Match specific conditions for determining desired state.

    :param plugin: Plugin class instance for Homie Automate.
    :param params: Parameters used to instantiate the class.
    """

    __plugin: 'HomiePlugin'

    __params: HomieParamsModel


    def __init__(
        self,
        plugin: 'HomiePlugin',
        params: HomieParamsModel,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__plugin = plugin
        self.__params = params

        self.__post__()


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        raise NotImplementedError


    @property
    def plugin(
        self,
    ) -> 'HomiePlugin':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__plugin


    @property
    def family(
        self,
    ) -> 'HomieFamily':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @property
    def kinds(
        self,
    ) -> list['HomiePluginKinds']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @property
    def params(
        self,
    ) -> HomieParamsModel:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    def occur(
        self,
        sitem: 'HomieStreamItem',
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        :param sitem: Item containing information for operation.
        :returns: Boolean indicating the conditional outcomes.
        """

        raise NotImplementedError


    def where(
        self,
        time: Time,
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        :param time: Time that will be used in the conditionals.
        :returns: Boolean indicating the conditional outcomes.
        """

        raise NotImplementedError



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import TYPE_CHECKING

from encommon.times import Time
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import save_text

from respx import MockRouter

from ..period import DriverBltnPeriod
from ...conftest import config_factory
from ...conftest import homie_factory

if TYPE_CHECKING:
    from ...homie import Homie
    from ...utils import TestBodies



def test_DriverBltnPeriod(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    desires = childs.desires


    planets = bodies.planets

    family = 'builtins'

    for planet in planets:

        desire = desires[
            f'{planet}_active']

        driver = (
            desire.wheres[3]
            .drivers[0])

        assert isinstance(
            driver, DriverBltnPeriod)


        attrs = lattrs(driver)

        assert attrs == [
            '_HomieDriver__plugin',
            '_HomieDriver__params']


        assert inrepr(
            'period.DriverBltnPeriod',
            driver)

        assert isinstance(
            hash(driver), int)

        assert instr(
            'period.DriverBltnPeriod',
            driver)


        driver.validate()

        assert driver.plugin

        assert driver.family == family

        assert driver.kinds == ['where']

        assert driver.params


        time = Time('05:00')

        assert driver.where(time)



def test_DriverBltnPeriod_cover(  # noqa: CFQ001
    tmp_path: Path,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    """

    samples = (
        tmp_path / 'homie')

    samples.mkdir()

    save_text(
        samples / 'test.yml',
        """
        desires:
         pytest:
          groups:
          - ganymede
          - halimede
          wheres:
          - builtins_period:
             start: '00:00'
             stop: '12:00'
          - builtins_period:
             start: '12:00'
          - builtins_period:
             days: Monday
             start: '00:00'
          - builtins_period:
             start: '12:00'
             stop: '17:00'
             tzname: US/Central
          - builtins_period:
             start: '00:00'
             stop: '07:00'
             tzname: US/Central
          - builtins_period:
             start: '19:00'
             tzname: US/Central
        """)  # noqa: LIT003

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock=respx_mock)


    childs = homie.childs

    desire = (
        childs.desires
        ['pytest'])


    driver = (
        desire.wheres[0]
        .drivers[0])

    asserts = [
        (Time('00:00'), True),
        (Time('11:00'), True),
        (Time('13:00'), False),
        (Time('23:00'), False)]

    assert all(
        driver.where(x) is y
        for x, y in asserts)


    driver = (
        desire.wheres[1]
        .drivers[0])

    asserts = [
        (Time('00:00'), False),
        (Time('11:00'), False),
        (Time('13:00'), True),
        (Time('23:00'), True)]

    assert all(
        driver.where(x) is y
        for x, y in asserts)


    driver = (
        desire.wheres[2]
        .drivers[0])

    assert driver.where(
        Time('Monday 12:00'))

    assert not driver.where(
        Time('Friday 12:00'))


    driver = (
        desire.wheres[3]
        .drivers[0])

    asserts = [
        (Time('16:30'), False),
        (Time('19:00'), True),
        (Time('23:30'), False)]

    assert all(
        driver.where(x) is y
        for x, y in asserts)


    driver = (
        desire.wheres[4]
        .drivers[0])

    asserts = [
        (Time('04:30'), False),
        (Time('05:30'), True),
        (Time('18:30'), False),
        (Time('23:30'), False)]

    assert all(
        driver.where(x) is y
        for x, y in asserts)


    driver = (
        desire.wheres[5]
        .drivers[0])

    asserts = [
        (Time('02:00'), True),
        (Time('07:00'), False)]

    assert all(
        driver.where(x) is y
        for x, y in asserts)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import DictStrAny

from .child import HomieChild
from ..models import HomieModels
from ...utils import InvalidParam
from ...utils import MultipleSource

if TYPE_CHECKING:
    from .origin import HomieOrigin
    from ..params import HomieDeviceParams



class HomieDevice(HomieChild):
    """
    Normalize the desired parameters with multiple products.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        homie = self.homie
        childs = homie.childs
        origins = childs.origins

        params = self.params

        _origin = params.origin
        _label = params.label
        _unique = params.unique


        if _origin not in origins:

            raise InvalidParam(
                child=self,
                param='origin',
                value=_origin,
                error='noexist')


        if (_unique is None
                and _label is None):

            raise InvalidParam(
                child=self,
                param='label',
                error='missing')


    @property
    def kind(
        self,
    ) -> Literal['device']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'device'


    @property
    def origin(
        self,
    ) -> 'HomieOrigin':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        homie = self.homie
        childs = homie.childs
        origins = childs.origins

        params = self.params
        origin = params.origin

        return origins[origin]


    @property
    def params(
        self,
    ) -> 'HomieDeviceParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            HomieModels
            .device())

        params = super().params

        assert isinstance(
            params, model)

        return params


    @property
    def source(
        self,
    ) -> Optional[DictStrAny]:
        """
        Return the content related to the item in parent system.

        :returns: Content related to the item in parent system.
        """

        homie = self.homie
        childs = homie.childs
        origins = childs.origins

        params = self.params

        _origin = params.origin
        _label = params.label
        _unique = params.unique

        origin = origins[_origin]

        try:

            return origin.source(
                kind='device',
                label=_label,
                unique=_unique)

        except MultipleSource:

            reason = (
                'multiple items '
                'match in source')

            homie.logger.log_e(
                base=self,
                name=self,
                status='multiple',
                reason=reason)

            return None


    @property
    def dumped(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        dumped = super().dumped

        origin = self.origin

        present = bool(self.source)

        return dumped | {
            'origin': origin.name,
            'present': present}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

if TYPE_CHECKING:
    from ...homie import Homie



def test_HomiePersist(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    persist = homie.persist


    attrs = lattrs(persist)

    assert attrs == [
        '_HomiePersist__homie',
        '_HomiePersist__connect',
        '_HomiePersist__locker',
        '_HomiePersist__sengine',
        '_HomiePersist__session']


    assert inrepr(
        'persist.HomiePersist',
        persist)

    assert isinstance(
        hash(persist), int)

    assert instr(
        'persist.HomiePersist',
        persist)


    persist.insert(
        unique='present',
        value=False)

    value = (
        persist
        .select('present'))

    assert value is False


    persist.delete(
        unique='present')

    value = (
        persist
        .select('present'))

    assert value is None


    persist.insert(
        unique='present',
        value=True,
        expire=-1)

    value = (
        persist
        .select('present'))

    assert value is None



def test_HomiePersist_cover(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    persist = homie.persist


    persist.insert(
        unique='present',
        value=None)

    value = (
        persist
        .select('present'))

    assert value is None


    persist.insert(
        unique='types',
        value=1)

    value = (
        persist
        .select('types'))

    assert value == 1


    persist.insert(
        unique='types',
        value=1.0)

    value = (
        persist
        .select('types'))

    assert value == 1.0


    persist.insert(
        unique='types',
        value='string')

    value = (
        persist
        .select('types'))

    assert value == 'string'


    persist.insert(
        unique='types',
        value=True)

    value = (
        persist
        .select('types'))

    assert value is True


    persist.insert(
        unique='types',
        value=None)

    value = (
        persist
        .select('types'))

    assert value is None



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING

from encommon.times import Time

from .models import BltnModels
from ..homie.plugins import HomieDriver
from ..utils import InvalidParam
from ..utils import UnexpectedCondition

if TYPE_CHECKING:
    from .params import DriverBltnStoreParams
    from ..homie.plugins import HomiePluginKinds



BltnStoreOperas = Literal[
    'present',
    'absent',
    'eq', 'neq',
    'lt', 'lte',
    'gt', 'gte']

_SINGLES = ['present', 'absent']



class DriverBltnStore(HomieDriver):
    """
    Match specific conditions for determining desired state.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        params = self.params

        _opera = params.operator
        _value = params.value


        if (_opera in _SINGLES
                and _value is not None):

            raise InvalidParam(
                param='operator',
                value=_value,
                error='invalid')


    @property
    def family(
        self,
    ) -> Literal['builtins']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'builtins'


    @property
    def kinds(
        self,
    ) -> list['HomiePluginKinds']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return ['where']


    @property
    def params(
        self,
    ) -> 'DriverBltnStoreParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            BltnModels
            .drivers()
            .store())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def where(  # noqa: CFQ004
        self,
        time: Time,
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        :param time: Time that will be used in the conditionals.
        :returns: Boolean indicating the conditional outcomes.
        """

        plugin = self.plugin
        homie = plugin.homie

        params = self.params
        unique = params.unique
        opera = params.operator
        value = params.value


        store = (
            homie.persist
            .select(unique))


        if opera == 'present':
            return store is not None

        if opera == 'absent':
            return store is None


        if (isinstance(value, int | float)
                and store is not None
                and store is not False
                and store is not True):

            store = float(store)

            if opera == 'lt':
                return store < value

            if opera == 'lte':
                return store <= value

            if opera == 'gt':
                return store > value

            if opera == 'gte':
                return store >= value


        if opera == 'neq':
            return store != value

        if opera == 'eq':
            return store == value


        raise UnexpectedCondition



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from encommon.types import DictStrAny
from encommon.types import getate
from encommon.types import setate



def surgeon(  # noqa: CFQ001, CFQ004
    fetch: DictStrAny,
    event: DictStrAny,
) -> bool:
    """
    Process the stream event updating the origin dictionary.

    :param fetch: Dictionary from the origin to be updated.
    :param event: Event from the stream for updating fetch.
    :returns: Boolean indicating whether or not was changed.
    """

    updated = False


    def _update_state() -> bool:

        _type = item['type']

        types = [
            'grouped_light',
            'light']

        if _type not in types:
            return False

        _update('on/on')

        return True


    def _update_color() -> bool:

        _type = item['type']

        types = [
            'grouped_light',
            'light']

        if _type not in types:
            return False

        _update('color/xy')

        return True


    def _update_level() -> bool:

        _type = item['type']

        types = [
            'grouped_light',
            'light']

        if _type not in types:
            return False

        _update(
            'dimming/brightness')

        return True


    def _update_scene() -> bool:

        _type = item['type']

        if _type != 'scene':
            return False

        _update('status/active')

        return True


    def _update_button() -> bool:

        _type = item['type']

        if _type != 'button':
            return False

        _update(
            'button/button_'
            'report/updated')

        _update(
            'button/button_'
            'report/event')

        _update(
            'button/last_event')

        return True


    def _update_contact() -> bool:

        _type = item['type']

        if _type != 'contact':
            return False

        _update(
            'contact_report'
            '/changed')

        _update(
            'contact_report'
            '/state')

        return True


    def _update_motion() -> bool:

        _type = item['type']

        if _type != 'motion':
            return False

        _update(
            'motion/motion')

        _update(
            'motion/motion_'
            'report/changed')

        _update(
            'motion/motion_'
            'report/event')

        _update(
            'motion/motion_valid')

        return True


    def _update(
        target: str,
    ) -> None:

        value = getate(event, target)

        if value is None:
            return None

        setate(item, target, value)


    unique = event['id']

    for item in fetch['data']:

        _unique = item['id']

        if unique != _unique:
            continue

        if _update_state():
            updated = True

        if _update_color():
            updated = True

        if _update_level():
            updated = True

        if _update_scene():
            updated = True

        if _update_button():
            updated = True

        if _update_contact():
            updated = True

        if _update_motion():
            updated = True


    return updated



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..regexp import DriverBltnRegexp

if TYPE_CHECKING:
    from ...homie import Homie
    from ...utils import TestBodies



def test_DriverBltnRegexp(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins
    aspires = childs.aspires


    assert homie.refresh()


    planets = bodies.planets

    family = 'builtins'

    event = {'foo': 'bar'}

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        aspire = aspires[
            f'{planet}_active']

        driver = (
            aspire.occurs[5]
            .drivers[0])

        assert isinstance(
            driver, DriverBltnRegexp)


        attrs = lattrs(driver)

        assert attrs == [
            '_HomieDriver__plugin',
            '_HomieDriver__params']


        assert inrepr(
            'regexp.DriverBltnRegexp',
            driver)

        assert isinstance(
            hash(driver), int)

        assert instr(
            'regexp.DriverBltnRegexp',
            driver)


        driver.validate()

        assert driver.plugin

        assert driver.family == family

        assert driver.kinds == ['occur']

        assert driver.params


        sitem = (
            origin
            .get_stream(event))

        assert driver.occur(sitem)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from argparse import ArgumentParser
from sys import argv as sys_argv
from typing import Optional

from encommon.types import DictStrAny
from encommon.utils import array_ansi
from encommon.utils import print_ansi

from ..homie import HomieConfig



def arguments(
    args: Optional[list[str]] = None,
) -> DictStrAny:
    """
    Construct arguments which are associated with the file.

    :param args: Override the source for the main arguments.
    :returns: Construct arguments from command line options.
    """

    parser = ArgumentParser()

    args = args or sys_argv[1:]


    parser.add_argument(
        '--config',
        required=True,
        help=(
            'complete or relative '
            'path to config file'))


    parser.add_argument(
        '--console',
        action='store_true',
        default=False,
        help=(
            'write log messages '
            'to standard output'))


    parser.add_argument(
        '--debug',
        action='store_true',
        default=False,
        help=(
            'increase logging level '
            'for standard output'))


    return vars(
        parser
        .parse_args(args))



def operation(
    # NOCVR
    config: HomieConfig,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param config: Primary class instance for configuration.
    """

    dumped = config.config

    print_ansi(
        array_ansi(dumped))



def execution(
    # NOCVR
    args: Optional[list[str]] = None,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param args: Override the source for the main arguments.
    """

    config = HomieConfig(
        arguments(args))

    config.logger.start()

    config.logger.log_i(
        base='execution/config',
        status='started')

    operation(config)

    config.logger.log_i(
        base='execution/config',
        status='stopped')

    config.logger.stop()



if __name__ == '__main__':
    execution()  # NOCVR



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.types import NCNone
from encommon.types import clsname
from encommon.types import sort_dict
from encommon.utils import array_ansi
from encommon.utils import print_ansi

from .addons import HomieAspired
from .addons import HomieDesired
from .addons import HomieLogger
from .addons import HomiePersist
from .childs import HomieChilds
from .models import HomieModels

if TYPE_CHECKING:
    from .common import HomiePrint
    from .common import HomieState
    from .childs import HomieDevice
    from .childs import HomieGroup
    from .childs import HomieScene
    from .config import HomieConfig
    from .params import HomieParams
    from .threads import HomieActionBase
    from .threads import HomieActionNode



class Homie:
    """
    Interact with supported devices to ensure desired state.

    :param config: Primary class instance for configuration.
    """

    __config: 'HomieConfig'

    __logger: HomieLogger
    __persist: HomiePersist

    __childs: HomieChilds


    def __init__(
        self,
        config: 'HomieConfig',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        config.logger.log_d(
            base=clsname(self),
            status='initial')

        self.__config = config

        self.__logger = (
            HomieLogger(self))

        self.__persist = (
            HomiePersist(self))

        self.__childs = (
            HomieChilds(self))

        self.childs.validate()

        self.__desired = (
            HomieDesired(self))

        self.__aspired = (
            HomieAspired(self))

        config.logger.log_d(
            base=clsname(self),
            status='created')


    @property
    def config(
        self,
    ) -> 'HomieConfig':
        """
        Return the Config instance containing the configuration.

        :returns: Config instance containing the configuration.
        """

        return self.__config


    @property
    def logger(
        self,
    ) -> HomieLogger:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__logger


    @property
    def persist(
        self,
    ) -> HomiePersist:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__persist


    @property
    def childs(
        self,
    ) -> HomieChilds:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__childs


    @property
    def params(
        self,
    ) -> 'HomieParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.config.params


    @property
    def dryrun(
        self,
    ) -> bool:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.params.dryrun


    @property
    def potent(
        self,
    ) -> bool:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.params.potent


    @property
    def desired(
        self,
    ) -> HomieDesired:
        """
        Return the related actions matching the provided event.

        :returns: Related actions matching the provided event.
        """

        return self.__desired


    @property
    def aspired(
        self,
    ) -> HomieAspired:
        """
        Return the related desired state for the desired groups.

        :returns: Related desired state for the desired groups.
        """

        return self.__aspired


    def refresh(
        self,
        timeout: Optional[int] = None,
    ) -> bool:
        """
        Refresh the cached information for the remote upstream.

        :param timeout: Timeout waiting for the server response.
        :returns: Boolean indicating the success of the refresh.
        """

        childs = self.childs

        refresh: set[bool] = set()

        origins = (
            childs.origins
            .values())


        for origin in origins:

            result = (
                origin
                .refresh(timeout))

            refresh.add(result)


        return all(refresh)


    @property
    def dumped(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        params = deepcopy(
            self.params.endumped)

        childs = deepcopy(
            self.childs.dumped)

        items = childs.items()

        for key, value in items:
            params[key] = value

        return params


    def printer(  # noqa: CFQ001,CFQ004
        self,
        source: 'HomiePrint',
        color: int = 6,
    ) -> None:
        """
        Print the contents for the object within Homie instance.

        :param source: Content which will be shown after header.
        :param color: Override the color used for box character.
        """

        params = self.params
        printer = params.printer

        paction = printer.action
        pupdate = printer.update
        pstream = printer.stream
        pdesire = printer.desire
        paspire = printer.aspire


        HomieActionItem = (
            HomieModels
            .action())

        HomieUpdateItem = (
            HomieModels
            .update())

        HomieStreamItem = (
            HomieModels
            .stream())

        HomieDesiredItem = (
            HomieModels
            .desired())

        HomieAspiredItem = (
            HomieModels
            .aspired())


        action = isinstance(
            source, HomieActionItem)

        if action and not paction:
            return NCNone


        update = isinstance(
            source, HomieUpdateItem)

        if update and not pupdate:
            return NCNone


        stream = isinstance(
            source, HomieStreamItem)

        if stream and not pstream:
            return NCNone


        desired = isinstance(
            source, HomieDesiredItem)

        if desired is True:

            if pdesire is False:
                return NCNone

            source = dict(
                source.__dict__)

            _target = source['target']
            _desire = source['desire']
            _scene = source['scene']

            target = (
                _target.name
                if _target is not None
                else None)

            desire = (
                _desire.name
                if _desire is not None
                else None)

            scene = (
                _scene.name
                if _scene is not None
                else None)

            source |= {
                'target': target,
                'desire': desire,
                'scene': scene}


        aspired = isinstance(
            source, HomieAspiredItem)

        if aspired is True:

            if paspire is False:
                return NCNone

            source = dict(
                source.__dict__)

            _target = source['target']
            _aspire = source['aspire']
            _scene = source['scene']

            target = (
                _target.name
                if _target is not None
                else None)

            aspire = (
                _aspire.name
                if _aspire is not None
                else None)

            scene = (
                _scene.name
                if _scene is not None
                else None)

            source |= {
                'target': target,
                'aspire': aspire,
                'scene': scene}


        line: str = '━'

        print_ansi(
            f'\n<c9{color}>┍'
            f'{line * 63}<c0>')

        dumped = array_ansi(
            source, indent=2)

        print(  # noqa: T201
            f'\n{dumped}')

        print_ansi(
            f'\n<c9{color}>┕'
            f'{line * 63}<c0>\n')


    def get_actions(  # noqa: CFQ002
        self,
        target: 'HomieActionNode',
        *,
        state: Optional['HomieState'] = None,
        color: Optional[str] = None,
        level: Optional[int] = None,
        scene: Optional['HomieScene'] = None,
        force: bool = False,
        change: bool = True,
    ) -> 'HomieActionBase':
        """
        Insert the new item containing information for operation.

        :param target: Device or group settings will be updated.
        :param state: Determine the state related to the target.
        :param color: Determine the color related to the target.
        :param level: Determine the level related to the target.
        :param scene: Determine the scene related to the target.
        :param force: Override the default for full idempotency.
        :param change: Determine whether the change is executed.
        """

        childs = self.childs
        devices = childs.devices

        aitems: 'HomieActionBase' = {}

        something = any([
            state is not None,
            color is not None,
            level is not None,
            scene is not None])

        if something is False:
            return aitems


        def _add_group(
            target: 'HomieGroup',
        ) -> None:

            origin = target.origin

            assert origin is not None

            aitem = (
                origin.get_action(
                    target=target,
                    state=state,
                    color=color,
                    level=level,
                    scene=scene))

            aitems[target] = aitem


        def _add_devices(
            target: 'HomieGroup',
        ) -> None:

            params = target.params
            names = params.devices

            assert names is not None

            for name in names:

                device = devices[name]

                _add_device(device)


        def _add_device(
            target: 'HomieDevice',
        ) -> None:

            origin = target.origin

            aitem = (
                origin.get_action(
                    target=target,
                    state=state,
                    color=color,
                    level=level,
                    scene=scene))

            aitems[target] = aitem


        if target.kind == 'group':

            params = target.params

            if params.devices:
                _add_devices(target)

            if target.origin:
                _add_group(target)


        if target.kind == 'device':
            _add_device(target)


        return sort_dict(aitems)


    def set_actions(
        self,
        aitems: 'HomieActionBase',
        force: bool = False,
        change: bool = True,
    ) -> bool:
        """
        Insert the new item containing information for operation.

        :param aitems: Constructed action objects for operation.
        :param force: Override the default for full idempotency.
        :param change: Determine whether the change is executed.
        :returns: Boolean indicating whether or not was changed.
        """

        changed: set[bool] = set()

        items = aitems.items()

        for target, aitem in items:

            assert target.origin

            origin = target.origin

            result = (
                origin.set_action(
                    target=target,
                    aitem=aitem,
                    force=force,
                    change=change))

            changed.add(result)

        return any(changed)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import save_text

from respx import MockRouter

from ..aspired import HomieAspiredItem
from ....conftest import config_factory
from ....conftest import homie_factory

if TYPE_CHECKING:
    from ...homie import Homie
    from ....utils import TestBodies



def test_HomieAspiredItem(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    childs = homie.childs
    devices = childs.devices
    groups = childs.groups
    aspires = childs.aspires

    device = devices[
        'jupiter_light1']

    group1 = groups['ganymede']
    group2 = groups['halimede']

    aspire = aspires[
        'jupiter_active']

    model = HomieAspiredItem


    item = model(
        group1, aspire)


    attrs = lattrs(item)

    assert attrs == [
        'target',
        'aspire',
        'state',
        'color',
        'level',
        'scene']


    assert item.target

    assert item.aspire

    assert item.state == 'poweron'

    assert not item.color

    assert item.level == 100

    assert item.scene


    item1 = model(
        group2, aspire)

    item2 = model(
        device, aspire)


    assert item < item1
    assert item < item2



def test_HomieAspired(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    origins = childs.origins


    aspired = homie.aspired


    attrs = lattrs(aspired)

    assert attrs == [
        '_HomieAspired__homie']


    assert inrepr(
        'aspired.HomieAspired',
        aspired)

    assert isinstance(
        hash(aspired), int)

    assert instr(
        'aspired.HomieAspired',
        aspired)


    assert homie.refresh()


    planets = bodies.planets

    total = 0

    event = {'foo': 'bar'}

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        sitem = (
            origin
            .get_stream(event))

        aitems = (
            aspired.items(sitem))

        total += len(aitems)

        if len(aitems) == 0:
            break

        assert len(aitems) == 4

    assert total == 4



def test_HomieAspired_cover(
    tmp_path: Path,
    respx_mock: MockRouter,
    replaces: DictStrAny,
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    :param replaces: Mapping of what to replace in samples.
    :param bodies: Locations and groups for use in testing.
    """

    samples = (
        tmp_path / 'homie')

    samples.mkdir()

    save_text(
        samples / 'test.yml',
        """
        aspires:
         pytest:
          pause: 1
          occurs:
          - builtins_regexp:
             patterns: '.*'
        """)  # noqa: LIT003

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock=respx_mock)


    childs = homie.childs
    origins = childs.origins
    aspired = homie.aspired


    assert homie.refresh()


    planets = bodies.planets

    event = {'foo': 'bar'}

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        sitem = (
            origin
            .get_stream(event))

        aitems = (
            aspired.items(sitem))

        if len(aitems) == 0:
            break

        assert len(aitems) == 4



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import dumps
from json import loads
from threading import Lock
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from encommon.times import Time
from encommon.times import unitime

from sqlalchemy import Column
from sqlalchemy import Numeric
from sqlalchemy import String
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Session
from sqlalchemy.orm import sessionmaker

if TYPE_CHECKING:
    from ..homie import Homie



HomiePersistValue = Optional[
    int | float | bool | str]

_PERSIST_VALUE = (
    int, float, bool, str)

HomiePersistExpire = Union[
    int, float, str]



class SQLBase(DeclarativeBase):
    """
    Some additional class that SQLAlchemy requires to work.
    """



class HomiePersistTable(SQLBase):
    """
    Schematic for the database operations using SQLAlchemy.

    .. note::
       Fields are not completely documented for this model.
    """

    unique = Column(
        String,
        primary_key=True,
        nullable=False)

    value = Column(
        String,
        nullable=False)

    expire = Column(
        Numeric,
        nullable=True)

    update = Column(
        Numeric,
        nullable=False)

    __tablename__ = 'persist'



class HomiePersist:
    """
    Store the persistent information in the key value store.

    :param homie: Primary class instance for Homie Automate.
    """

    __connect: str
    __locker: Lock

    __sengine: Engine
    __session: (
        # pylint: disable=unsubscriptable-object
        sessionmaker[Session])


    def __init__(
        self,
        homie: 'Homie',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__homie = homie

        params = homie.params

        self.__connect = (
            params.database)

        self.__locker = Lock()

        self.__build_engine()


    def __build_engine(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        path = self.__connect

        sengine = (
            create_engine(path))

        (SQLBase.metadata
         .create_all(sengine))

        session = (
            sessionmaker(sengine))

        self.__sengine = sengine
        self.__session = session


    def insert(
        self,
        unique: str,
        value: HomiePersistValue,
        expire: HomiePersistExpire = '30d',
    ) -> None:
        """
        Insert the value within the persistent key value store.

        .. note::
           This will replace existing record with primary key.

        :param unique: Unique identifier from within the table.
        :param value: Value Which will be stored within record.
        :param expire: Optional time in seconds for expiration.
        """

        expire = unitime(expire)

        assert isinstance(expire, int)

        update = Time().spoch
        expire = update + expire


        if value is None:
            self.delete(unique)
            return None


        sess = self.__session()
        lock = self.__locker

        with lock, sess as session:

            session.merge(
                HomiePersistTable(
                    unique=unique,
                    value=dumps(value),
                    expire=expire,
                    update=update))

            session.commit()


    def select(
        self,
        unique: str,
    ) -> HomiePersistValue:
        """
        Return the value from within persistent key value store.

        :param unique: Unique identifier from within the table.
        :returns: Value from within persistent key value store.
        """

        sess = self.__session()
        lock = self.__locker

        table = HomiePersistTable
        field = table.unique

        self.expire()

        with lock, sess as session:

            query = (
                session.query(table)
                .filter(field == unique))

            record = query.first()

            if record is None:
                return None

            value = str(record.value)

        value = loads(value)

        assert isinstance(
            value, _PERSIST_VALUE)

        return value


    def delete(
        self,
        unique: str,
    ) -> None:
        """
        Delete the value within the persistent key value store.

        :param unique: Unique identifier from within the table.
        """

        sess = self.__session()
        lock = self.__locker

        table = HomiePersistTable
        field = table.unique

        with lock, sess as session:

            query = (
                session.query(table)
                .filter(field == unique))

            record = query.first()

            if record is None:
                return None

            session.delete(record)

            session.commit()


    def expire(
        self,
    ) -> None:
        """
        Remove the expired persistent key values from the table.
        """

        table = HomiePersistTable
        expire = table.expire

        sess = self.__session()
        lock = self.__locker

        with lock, sess as session:

            now = Time().spoch

            (session.query(table)
             .filter(expire < now)
             .delete())

            session.commit()



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.types import merge_dicts

from .child import HomieChild
from ..common import HomieStage
from ..models import HomieModels
from ...utils import InvalidParam
from ...utils import MultipleSource

if TYPE_CHECKING:
    from .origin import HomieOrigin
    from .device import HomieDevice
    from ..threads import HomieActionNode
    from ..params import HomieSceneParams



class HomieScene(HomieChild):
    """
    Normalize the desired parameters with multiple products.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        homie = self.homie
        childs = homie.childs
        devices = childs.devices

        params = self.params

        _label = params.label
        _stage = params.stage
        _devices = params.devices


        if (_stage is None
                and _label is None):

            raise InvalidParam(
                child=self,
                error='minimal')


        for value in _devices or []:

            if value in devices:
                continue

            raise InvalidParam(
                child=self,
                param='devices',
                value=value,
                error='noexist')


    @property
    def kind(
        self,
    ) -> Literal['scene']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'scene'


    @property
    def params(
        self,
    ) -> 'HomieSceneParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            HomieModels
            .scene())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def source(
        self,
        origin: 'HomieOrigin',
        relate: Optional['HomieActionNode'] = None,
    ) -> Optional[DictStrAny]:
        """
        Return the content related to the item in parent system.

        :param origin: Child class instance for Homie Automate.
        :param relate: Child class instance for Homie Automate.
        :returns: Content related to the item in parent system.
        """

        homie = self.homie
        params = self.params

        _label = params.label

        try:

            return origin.source(
                kind='scene',
                label=_label,
                relate=relate)

        except MultipleSource:

            reason = (
                'multiple items '
                'match in source')

            homie.logger.log_e(
                base=self,
                name=self,
                status='multiple',
                reason=reason)

            return None


    def stage(
        self,
        device: Optional['HomieDevice'] = None,
    ) -> HomieStage:
        """
        Return the configuration for the device or the defaults.

        :param device: Child class instance for Homie Automate.
        :returns: Configuration for the device or the defaults.
        """

        params = self.params
        stage = params.stage
        devices = params.devices

        pruned = (
            stage
            .enpruned)

        pruned = deepcopy(pruned)


        def _override() -> None:

            assert device and devices

            config = devices[
                device.name]

            _pruned = (
                config
                .enpruned)

            _pruned = deepcopy(_pruned)

            merge_dicts(
                dict1=pruned,
                dict2=_pruned,
                force=True)


        if device and devices:

            name = device.name

            if name in devices:
                _override()


        return HomieStage(**pruned)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

if TYPE_CHECKING:
    from ...homie import Homie
    from ....utils import TestBodies



def test_HomieWhere(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    desires = childs.desires


    planets = bodies.planets

    for planet in planets:

        desire = desires[
            f'{planet}_active']

        wheres = desire.wheres


        plugin = wheres[0]


        attrs = lattrs(plugin)

        assert attrs == [
            '_HomiePlugin__homie',
            '_HomiePlugin__params',
            '_HomiePlugin__drivers']


        assert inrepr(
            'where.HomieWhere',
            plugin)

        assert isinstance(
            hash(plugin), int)

        assert instr(
            'where.HomieWhere',
            plugin)


        plugin.validate()

        assert plugin.homie

        assert plugin.family == 'builtins'

        assert plugin.kind == 'where'

        assert plugin.params

        assert len(plugin.drivers) == 1



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING
from typing import get_args

from encommon.times import Time

from .helpers import phue_changed
from ..models import PhueModels
from ...homie.plugins import HomieDriver
from ...utils import InvalidParam

if TYPE_CHECKING:
    from ..params import DriverPhueChangeParams
    from ...homie.plugins import HomiePluginKinds



PhueChangeSensor = Literal[
    'button1',
    'button2',
    'button3',
    'button4',
    'contact',
    'motion']

_SENSORS = list(
    get_args(PhueChangeSensor))



class DriverPhueChange(HomieDriver):
    """
    Match specific conditions for determining desired state.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        plugin = self.plugin
        family = self.family
        homie = plugin.homie
        childs = homie.childs
        devices = childs.devices
        params = self.params

        _devices = params.devices


        for name in _devices:

            if name in devices:
                continue

            raise InvalidParam(
                param='devices',
                value=name,
                error='noexist')


        for name in _devices:

            device = devices[name]

            _family = device.family

            if _family == family:
                continue

            raise InvalidParam(
                param='devices',
                value=name,
                error='invalid')


    @property
    def family(
        self,
    ) -> Literal['philips']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'philips'


    @property
    def kinds(
        self,
    ) -> list['HomiePluginKinds']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return ['where']


    @property
    def params(
        self,
    ) -> 'DriverPhueChangeParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            PhueModels
            .drivers()
            .change())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def where(  # noqa: CFQ004
        self,
        time: Time,
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        :param time: Time that will be used in the conditionals.
        :returns: Boolean indicating the conditional outcomes.
        """

        plugin = self.plugin
        homie = plugin.homie
        childs = homie.childs
        devices = childs.devices

        params = self.params
        targets = params.devices
        since = params.since
        sensors = (
            params.sensors
            or _SENSORS)

        wheres: list[bool] = []


        def _process() -> None:

            if key not in sensors:
                return None

            if value is None:
                return None

            _since = time - value

            wheres.append(
                _since >= since)


        for target in targets:

            source = (
                devices[target]
                .source)

            if source is None:
                continue

            items = (
                phue_changed(source)
                .items())

            for key, value in items:
                _process()


        if len(wheres) >= 1:
            return any(wheres)

        return True



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Literal
from typing import get_args

from encommon.times import Time



STARTED = Time('now')



Planets = Literal[
    'jupiter', 'neptune']

PLANETS = set(
    get_args(Planets))

_PLANETS = tuple[Planets, ...]



Moons = Literal[
    'ganymede', 'halimede']

MOONS = set(
    get_args(Moons))

MoonPair = tuple[
    Planets, Moons]

_MOONS = tuple[MoonPair, ...]



@dataclass(frozen=True)
class TestTimes:
    """
    Contain the times used within various tests in project.
    """

    start: Time = STARTED.shift('-1w@h')
    middle: Time = STARTED.shift('-1d@h')
    final: Time = STARTED.shift('-1h@h')
    current: Time = STARTED.shift('-1m')



@dataclass(frozen=True)
class TestBodies:
    """
    Contain the planets used to reference test sample sets.
    """

    planets: _PLANETS = tuple(PLANETS)

    moons: _MOONS = (
        ('jupiter', 'ganymede'),
        ('neptune', 'halimede'))



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.colors import Color
from encommon.types import DictStrAny
from encommon.types import NCFalse
from encommon.types import NCNone
from encommon.types import getate
from encommon.types import setate
from encommon.types import strplwr

from httpx import Response

from .device import PhueDevice
from ..utils import Idempotent

if TYPE_CHECKING:
    from .origin import PhueOrigin
    from ..homie.common import HomieKinds
    from ..homie.common import HomieState
    from ..homie.childs import HomieScene
    from ..homie.threads import HomieActionNode



PhueFetch = DictStrAny
PhueMerge = dict[str, PhueFetch]



_RESPONSE = Optional[
    Response | Literal[True]]

_APROPS = Literal[
    'state',
    'color',
    'level']

_ATYPES = Literal[
    'scene',
    'grouped_light',
    'light']

_ASTATES = Literal[
    'on', 'off',
    'active']

_AVALUES = Optional[
    _ASTATES
    | int | Color]



def merge_fetch(
    fetch: PhueFetch,
) -> PhueMerge:
    """
    Return the content related to the item in parent system.

    :param fetch: Dictionary from the origin to be updated.
    :returns: Content related to the item in parent system.
    """

    fetch = deepcopy(fetch)

    source = {
        x['id']: x for x
        in fetch['data']}

    origin = deepcopy(source)


    def _enrichment() -> None:

        rtype = item['rtype']
        rid = item['rid']

        if 'taurus_' in rtype:
            return None

        item['_source'] = (
            origin[rid])


    items1 = source.items()

    for key, value in items1:

        if 'services' not in value:
            continue

        items2 = value['services']

        for item in items2:
            _enrichment()


    _source = sorted(
        source.items())

    return dict(_source)



def merge_find(  # noqa: CFQ001,CFQ004
    merge: PhueMerge,
    kind: Optional['HomieKinds'] = None,
    unique: Optional[str] = None,
    label: Optional[str] = None,
    relate: Optional['HomieActionNode'] = None,
) -> list[DictStrAny]:
    """
    Return the content related to the item in parent system.

    :param merge: Dictionary from the origin to be updated.
    :param kind: Which kind of Homie object will be located.
    :param unique: Unique identifier within parents system.
    :param label: Friendly name or label within the parent.
    :param relate: Child class instance for Homie Automate.
    :returns: Content related to the item in parent system.
    """

    assert unique or label

    if label is not None:
        label = strplwr(label)

    if unique is not None:
        unique = strplwr(unique)

    found: list[DictStrAny] = []


    def _match_owner() -> bool:

        if relate is None:
            return True

        _source = relate.source

        if _source is None:
            return NCFalse

        _relate = (
            _source['id']
            if relate is not None
            else None)

        owner = getate(
            value, 'owner/rid')

        group = getate(
            value, 'group/rid')

        match = owner or group

        return (
            False  # noqa: SIM211
            if _relate != match
            else True)


    def _match_kind() -> bool:

        if kind is None:
            return True

        _kind = value['type']

        assert isinstance(_kind, str)

        if _kind == 'room':
            _kind = 'group'

        if _kind == 'zone':
            _kind = 'group'

        return kind == _kind


    def _match_unique() -> bool:

        if unique is None:
            return True

        _unique = strplwr(
            value['id'])

        return unique == _unique


    def _match_label() -> bool:

        if label is None:
            return True

        name = getate(
            value,
            'metadata/name')

        if name is None:
            return False

        _label = strplwr(name)

        return label == _label


    values = merge.values()

    for value in values:

        if not _match_owner():
            continue

        if not _match_kind():
            continue

        if not _match_unique():
            continue

        if not _match_label():
            continue

        found.append(value)


    return found



def request_action(  # noqa: CFQ001,CFQ002,CFQ004
    origin: 'PhueOrigin',
    target: 'HomieActionNode',
    *,
    state: Optional['HomieState'] = None,
    color: Optional[str | Color] = None,
    level: Optional[int] = None,
    scene: Optional['HomieScene'] = None,
    force: bool = False,
    change: bool = True,
    timeout: Optional[int] = None,
) -> None:
    """
    Perform the provided action with specified Homie target.

    :param origin: Child class instance for Homie Automate.
    :param target: Device or group settings will be updated.
    :param state: Determine the state related to the target.
    :param color: Determine the color related to the target.
    :param level: Determine the level related to the target.
    :param scene: Determine the scene related to the target.
    :param force: Override the default for full idempotency.
    :param change: Determine whether the change is executed.
    :param timeout: Timeout waiting for the server response.
    """

    homie = origin.homie
    kind = target.kind

    request = action_request

    changed: set[bool] = set()

    source = target.source


    if (scene is not None
            and kind == 'device'):

        assert isinstance(
            target, PhueDevice)

        stage = (
            scene
            .stage(target))

        if state is None:
            state = stage.state

        if color is None:
            color = stage.color

        if level is None:
            level = stage.level


    elif (scene is not None
            and kind == 'group'):

        source = scene.source(
            origin, target)


    assert source is not None


    _state: _ASTATES = (
        'off'
        if state == 'nopower'
        else 'on')


    if isinstance(color, str):
        color = Color(color)


    def _set_scene() -> None:

        unique = source['id']

        response = request(
            origin=origin,
            type='scene',
            unique=unique,
            state='active',
            force=force,
            change=change,
            timeout=timeout)

        if response is None:
            return None

        changed.add(True)

        if change is False:
            return None

        if homie.dryrun:
            return None

        assert isinstance(
            response, Response)

        (response
         .raise_for_status())


    def _set_target(  # noqa: CFQ004
        item: DictStrAny,
    ) -> None:

        rtype = item['rtype']
        rid = item['rid']

        itype: _ATYPES = (
            'grouped_light'
            if kind == 'group'
            else 'light')

        if rtype != itype:
            return None

        response = request(
            origin=origin,
            type=itype,
            unique=rid,
            state=(
                _state
                if state is not None
                else None),
            color=color,
            level=level,
            force=force,
            change=change,
            timeout=timeout)

        if response is None:
            return None

        changed.add(True)

        if change is False:
            return None

        if homie.dryrun:
            return NCNone

        assert isinstance(
            response, Response)

        (response
         .raise_for_status())


    def _set_targets() -> None:

        services = (
            source['services'])

        for item in services:
            _set_target(item)


    if kind == 'device':
        _set_targets()

    elif (kind == 'group'
            and scene is None):
        _set_targets()

    elif kind == 'group':
        _set_scene()


    if not any(changed):
        raise Idempotent



def action_request(  # noqa: CFQ001,CFQ002,CFQ004
    origin: 'PhueOrigin',
    type: _ATYPES,
    unique: str,
    *,
    state: Optional[_ASTATES] = None,
    color: Optional[Color] = None,
    level: Optional[int] = None,
    force: bool = False,
    change: bool = True,
    timeout: Optional[int] = None,
) -> _RESPONSE:
    """
    Request to execute the action on target from the bridge.

    :param origin: Child class instance for Homie Automate.
    :param type: Which type of target to perform the action.
    :param unique: Unique identifier within parents system.
    :param state: Determine the state related to the target.
    :param color: Determine the color related to the target.
    :param level: Determine the level related to the target.
    :param force: Override the default for full idempotency.
    :param change: Determine whether the change is executed.
    :param timeout: Timeout waiting for the server response.
    :returns: Response from upstream request to the server.
    """

    bridge = origin.bridge
    request = bridge.request
    homie = origin.homie

    merge = origin.source(
        unique=unique)

    assert merge is not None


    data: DictStrAny = {}

    path = (
        f'resource/{type}'
        f'/{unique}')


    def _set_scene() -> None:

        if state != 'active':
            return None

        key = 'recall/action'
        value = 'active'

        crrnt = getate(
            merge, 'status/active')

        if (crrnt != 'inactive'
                and not force):
            return None

        setate(data, key, value)


    _set_scene()


    def _set_color() -> None:

        if color is None:
            return None

        key = 'color/xy'

        value = {
            'x': color.xy[0],
            'y': color.xy[1]}

        crrnt = getate(merge, key)

        if (crrnt == value
                and not force):
            return None

        setate(data, key, value)


    _set_color()


    def _set_level() -> None:

        if level is None:
            return None

        key = (
            'dimming/'
            'brightness')

        crrnt = int(
            getate(merge, key))

        if (crrnt == level
                and not force):
            return None

        setate(data, key, level)


    _set_level()


    def _set_state() -> None:

        states = ['on', 'off']

        if state not in states:
            return None

        key = 'on/on'
        value = state == 'on'

        crrnt = getate(merge, key)

        if (crrnt == value
                and not force):
            return None

        setate(data, key, value)


    _set_state()


    if len(data) == 0:
        return None

    if change is False:
        return True

    if homie.dryrun:
        return True


    return request(
        method='put',
        path=path,
        json=data,
        timeout=timeout)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from encommon.colors import Color
from encommon.types import DictStrAny

from .child import HomieChild
from ..addons import HomieQueue
from ..models import HomieModels
from ...utils import InvalidParam

if TYPE_CHECKING:
    from .scene import HomieScene
    from ..common import HomieFamily
    from ..common import HomieKinds
    from ..common import HomieState
    from ..params import HomieOriginParams
    from ..threads import HomieActionItem
    from ..threads import HomieActionNode
    from ..threads import HomieUpdateBase
    from ..threads import HomieUpdateItem
    from ..threads import HomieStreamItem



class HomieOrigin(HomieChild):
    """
    Normalize the desired parameters with multiple products.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        params = self.params

        _hubitat = params.hubitat
        _philips = params.philips
        _ubiquiti = params.ubiquiti


        enabled = [
            1 if _hubitat else 0,
            1 if _philips else 0,
            1 if _ubiquiti else 0]

        if sum(enabled) != 1:

            raise InvalidParam(
                child=self,
                error='minimal')


    @property
    def family(
        self,
    ) -> 'HomieFamily':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @property
    def kind(
        self,
    ) -> Literal['origin']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'origin'


    @property
    def params(
        self,
    ) -> 'HomieOriginParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            HomieModels
            .origin())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def refresh(
        self,
        timeout: Optional[int] = None,
    ) -> bool:
        """
        Refresh the cached information for the remote upstream.

        :param timeout: Timeout waiting for the server response.
        :returns: Boolean indicating the success of the refresh.
        """

        raise NotImplementedError


    def source(
        self,
        kind: Optional['HomieKinds'] = None,
        unique: Optional[str] = None,
        label: Optional[str] = None,
        relate: Optional['HomieActionNode'] = None,
    ) -> Optional[DictStrAny]:
        """
        Return the content related to the item in parent system.

        :param kind: Which kind of Homie object will be located.
        :param unique: Unique identifier within parents system.
        :param label: Friendly name or label within the parent.
        :param relate: Child class instance for Homie Automate.
        :returns: Content related to the item in parent system.
        """

        raise NotImplementedError


    @property
    def dumped(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        dumped = super().dumped

        present = bool(self.source)

        return dumped | {
            'present': present}


    def get_action(
        self,
        target: 'HomieActionNode',
        *,
        state: Optional['HomieState'] = None,
        color: Optional[str | Color] = None,
        level: Optional[int] = None,
        scene: Optional[Union[str, 'HomieScene']] = None,
    ) -> 'HomieActionItem':
        """
        Return the new item containing information for operation.

        :param target: Device or group settings will be updated.
        :param state: Determine the state related to the target.
        :param color: Determine the color related to the target.
        :param level: Determine the level related to the target.
        :param scene: Determine the scene related to the target.
        :returns: New item containing information for operation.
        """

        raise NotImplementedError


    def set_action(
        self,
        target: 'HomieActionNode',
        aitem: 'HomieActionItem',
        force: bool = False,
        change: bool = True,
        timeout: Optional[int] = None,
    ) -> bool:
        """
        Perform the provided action with specified Homie target.

        :param target: Device or group settings will be updated.
        :param aitem: Item containing information for operation.
        :param force: Override the default for full idempotency.
        :param change: Determine whether the change is executed.
        :param timeout: Timeout waiting for the server response.
        :returns: Boolean indicating whether or not was changed.
        """

        raise NotImplementedError


    def put_action(  # noqa: CFQ002
        self,
        aqueue: HomieQueue['HomieActionItem'],
        target: 'HomieActionNode',
        *,
        state: Optional['HomieState'] = None,
        color: Optional[str | Color] = None,
        level: Optional[int] = None,
        scene: Optional[Union[str, 'HomieScene']] = None,
    ) -> None:
        """
        Insert the new item containing information for operation.

        :param aqueue: Queue instance where the item is received.
        :param target: Device or group settings will be updated.
        :param state: Determine the state related to the target.
        :param color: Determine the color related to the target.
        :param level: Determine the level related to the target.
        :param scene: Determine the scene related to the target.
        """

        aitem = (
            self.get_action(
                target=target,
                state=state,
                color=color,
                level=level,
                scene=scene))

        aqueue.put(aitem)


    def get_update(
        self,
    ) -> 'HomieUpdateItem':
        """
        Return the new item containing information for operation.

        :returns: New item containing information for operation.
        """

        raise NotImplementedError


    def set_update(
        self,
        uitem: 'HomieUpdateBase',
    ) -> bool:
        """
        Replace or update internal configuration with provided.

        :param uitem: Item containing information for operation.
        :returns: Boolean indicating whether or not was changed.
        """

        raise NotImplementedError


    def put_update(
        self,
        uqueue: HomieQueue['HomieUpdateItem'],
    ) -> None:
        """
        Insert the new item containing information for operation.

        :param uqueue: Queue instance where the item is received.
        """

        uitem = self.get_update()

        uqueue.put(uitem)


    def get_stream(
        self,
        event: DictStrAny,
    ) -> 'HomieStreamItem':
        """
        Return the new item containing information for operation.

        :param event: Event from the stream for creating object.
        :returns: New item containing information for operation.
        """

        raise NotImplementedError


    def put_stream(
        self,
        squeue: HomieQueue['HomieStreamItem'],
        event: DictStrAny,
    ) -> None:
        """
        Insert the new item containing information for operation.

        :param squeue: Queue instance where the item is received.
        :param event: Event from the stream for creating object.
        """

        stream = (
            self
            .get_stream(event))

        squeue.put(stream)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .device import UbiqDevice
from .models import UbiqModels
from .origin import UbiqOrigin
from .plugins import DriverUbiqClient
from .update import UbiqUpdate



__all__ = [
    'UbiqOrigin',
    'UbiqDevice',
    'UbiqUpdate',
    'UbiqModels',
    'DriverUbiqClient']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING

from encommon.times import Time

from .helpers import ubiq_latest
from ..models import UbiqModels
from ...homie.plugins import HomieDriver
from ...utils import InvalidParam

if TYPE_CHECKING:
    from ..params import DriverUbiqClientParams
    from ...homie.plugins import HomiePluginKinds



class DriverUbiqClient(HomieDriver):
    """
    Match specific conditions for determining desired state.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        plugin = self.plugin
        family = self.family
        homie = plugin.homie
        childs = homie.childs
        devices = childs.devices
        params = self.params

        _clients = params.clients


        for name in _clients:

            if name in devices:
                continue

            raise InvalidParam(
                param='clients',
                value=name,
                error='noexist')


        for name in _clients:

            device = devices[name]

            _family = device.family

            if _family == family:
                continue

            raise InvalidParam(
                param='clients',
                value=name,
                error='invalid')


    @property
    def family(
        self,
    ) -> Literal['ubiquiti']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'ubiquiti'


    @property
    def kinds(
        self,
    ) -> list['HomiePluginKinds']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return ['where']


    @property
    def params(
        self,
    ) -> 'DriverUbiqClientParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            UbiqModels
            .drivers()
            .client())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def where(
        self,
        time: Time,
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        :param time: Time that will be used in the conditionals.
        :returns: Boolean indicating the conditional outcomes.
        """

        plugin = self.plugin
        homie = plugin.homie
        childs = homie.childs
        devices = childs.devices

        params = self.params
        targets = params.clients
        since = params.since

        wheres: list[bool] = []


        for target in targets:

            source = (
                devices[target]
                .source)

            if source is None:
                continue

            value = ubiq_latest(source)

            _since = time - value

            wheres.append(
                _since >= since)


        if len(wheres) >= 1:
            return any(wheres)

        return True



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .aspire import HomieAspireParams
from .child import HomieChildParams
from .common import HomieParamsModel
from .desire import HomieDesireParams
from .device import HomieDeviceParams
from .group import HomieGroupParams
from .homie import HomieParams
from .homie import HomiePrinterParams
from .occur import HomieOccurParams
from .origin import HomieOriginParams
from .plugin import HomiePluginParams
from .scene import HomieSceneParams
from .service import HomieServiceParams
from .store import HomieStoreParams
from .where import HomieWhereParams



__all__ = [
    'HomieParams',
    'HomieParamsModel',
    'HomiePrinterParams',
    'HomieChildParams',
    'HomieOriginParams',
    'HomieDeviceParams',
    'HomieGroupParams',
    'HomieSceneParams',
    'HomieAspireParams',
    'HomieDesireParams',
    'HomieServiceParams',
    'HomiePluginParams',
    'HomieStoreParams',
    'HomieOccurParams',
    'HomieWhereParams']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inlist
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from pytest import raises

from ..stream import HomieStreamItem

if TYPE_CHECKING:
    from ...service import HomieService
    from ....utils import TestBodies



def test_HomieStreamItem(
    service: 'HomieService',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    :param bodies: Locations and groups for use in testing.
    """

    homie = service.homie
    childs = homie.childs
    origins = childs.origins

    model = HomieStreamItem


    planets = bodies.planets

    event = {'foo': 'bar'}

    for planet in planets:

        origin = origins[
            f'{planet}_philips']


        item = model(
            origin, event)


        attrs = lattrs(item)

        assert attrs == [
            'event',
            'origin',
            'time']


        assert inrepr(
            'HomieStreamItem',
            item)

        with raises(TypeError):
            hash(item)

        assert instr(
            'HomieStreamItem',
            item)


        assert item.time.since > 0

        assert item.origin == origin.name

        assert item.event == event



def test_HomieStream(
    service: 'HomieService',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    :param bodies: Locations and groups for use in testing.
    """

    member = service.streams
    threads = member.threads


    planets = bodies.planets

    for planet in planets:

        thread = threads[
            f'{planet}_philips']


        attrs = lattrs(thread)

        # Inherits Thread class

        assert inlist(
            '_HomieThread__member',
            attrs)

        assert inlist(
            '_HomieThread__origin',
            attrs)


        assert inrepr(
            'Stream(PhueStream',
            thread)

        assert isinstance(
            hash(thread), int)

        assert instr(
            'Stream(PhueStream',
            thread)


        assert thread.homie

        assert thread.member

        assert thread.origin

        assert thread.aqueue

        assert thread.uqueue

        assert thread.squeue

        assert not thread.congest

        assert not thread.enqueue



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Optional

from pydantic import Field

from .common import HomieParamsModel
from ..common import HomieState



class HomieStageParams(HomieParamsModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    state: Annotated[
        Optional[HomieState],
        Field(None,
              description='Setting for the light state')]

    color: Annotated[
        Optional[str],
        Field(None,
              description='Setting for the light color',
              min_length=6,
              max_length=6)]

    level: Annotated[
        Optional[int],
        Field(None,
              description='Setting for the light level',
              ge=0, le=100)]



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import clsname

if TYPE_CHECKING:
    from ..homie.childs import HomieChild



_ERROR = Literal[
    'minimal',
    'invalid',
    'missing',
    'noexist']



class InvalidParam(Exception):
    """
    Exception for after invalid parameters are encountered.

    .. note::
       Similar or identical to classes in other projects.

    :param error: Simple code describing the invalid error.
    :param about: Additional information for the exception.
    :param child: Child class instance for Homie Automate.
    :param param: Name of the parameter which is not valid.
    :param value: Value if any specified for the parameter.
    """

    error: _ERROR
    about: Optional[str] = None
    child: Optional['HomieChild'] = None
    param: Optional[str] = None
    value: Optional[Any] = None


    def __init__(
        self,
        error: _ERROR,
        about: Optional[str] = None,
        *,
        child: Optional['HomieChild'] = None,
        param: Optional[str] = None,
        value: Optional[Any] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        message = (
            f'Error ({error})')


        if param is not None:
            message += (
                f' param ({param})')

        if value is not None:
            message += (
                f' value ({value})')


        if child is not None:

            name = (
                f'{clsname(child)}'
                f'/{child.name}')

            message += (
                f' child ({name})')


        if about is not None:
            message += (
                f' ({about})')


        self.error = error
        self.about = about
        self.child = child
        self.param = param
        self.value = value

        super().__init__(message)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING
from typing import Type

if TYPE_CHECKING:
    from .params import DriverUbiqClientParams
    from .params import UbiqOriginParams
    from .update import UbiqUpdateItem



class UbiqDriverModels:
    """
    Return the class object that was imported within method.
    """


    @classmethod
    def client(
        cls,
    ) -> Type['DriverUbiqClientParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            DriverUbiqClientParams)

        return DriverUbiqClientParams



class UbiqModels:
    """
    Return the class object that was imported within method.
    """


    @classmethod
    def origin(
        cls,
    ) -> Type['UbiqOriginParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .params import (
            UbiqOriginParams)

        return UbiqOriginParams


    @classmethod
    def update(
        cls,
    ) -> Type['UbiqUpdateItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from .update import (
            UbiqUpdateItem)

        return UbiqUpdateItem


    @classmethod
    def drivers(
        cls,
    ) -> Type['UbiqDriverModels']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        return UbiqDriverModels



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Optional

from pydantic import Field

from .plugin import HomiePluginParams
from ...builtins.params import DriverBltnRegexpParams
from ...philips.params import DriverPhueButtonParams
from ...philips.params import DriverPhueContactParams
from ...philips.params import DriverPhueMotionParams
from ...philips.params import DriverPhueSceneParams



class HomieOccurParams(HomiePluginParams, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    builtins_regexp: Annotated[
        Optional[DriverBltnRegexpParams],
        Field(None,
              description='Plugin for the operations')]

    philips_button: Annotated[
        Optional[DriverPhueButtonParams],
        Field(None,
              description='Plugin for the operations')]

    philips_contact: Annotated[
        Optional[DriverPhueContactParams],
        Field(None,
              description='Plugin for the operations')]

    philips_motion: Annotated[
        Optional[DriverPhueMotionParams],
        Field(None,
              description='Plugin for the operations')]

    philips_scene: Annotated[
        Optional[DriverPhueSceneParams],
        Field(None,
              description='Plugin for the operations')]



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal

from ..homie.childs import HomieDevice



class PhueDevice(HomieDevice):
    """
    Contain the relevant attributes from the related source.
    """


    @property
    def family(
        self,
    ) -> Literal['philips']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'philips'



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .action import HubiAction
from .device import HubiDevice
from .models import HubiModels
from .origin import HubiOrigin
from .update import HubiUpdate



__all__ = [
    'HubiOrigin',
    'HubiDevice',
    'HubiAction',
    'HubiUpdate',
    'HubiModels']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING
from typing import get_args

from encommon.types import NCFalse
from encommon.types import getate

from ..models import PhueModels
from ..stream import PhueStreamItem
from ...homie.plugins import HomieDriver
from ...utils import InvalidParam

if TYPE_CHECKING:
    from ..params import DriverPhueSceneParams
    from ...homie.plugins import HomiePluginKinds
    from ...homie.threads import HomieStreamItem



PhueSceneState = Literal[
    'active',
    'inactive']

_STATES = list(
    get_args(PhueSceneState))



class DriverPhueScene(HomieDriver):
    """
    Match specific conditions for determining desired state.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        plugin = self.plugin
        homie = plugin.homie
        childs = homie.childs
        groups = childs.groups
        scenes = childs.scenes
        params = self.params

        _scene = params.scene
        _group = params.group


        if _scene not in scenes:

            raise InvalidParam(
                param='scene',
                value=_scene,
                error='noexist')


        if _group not in groups:

            raise InvalidParam(
                param='group',
                value=_group,
                error='noexist')


    @property
    def family(
        self,
    ) -> Literal['philips']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'philips'


    @property
    def kinds(
        self,
    ) -> list['HomiePluginKinds']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return ['occur']


    @property
    def params(
        self,
    ) -> 'DriverPhueSceneParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            PhueModels
            .drivers()
            .scene())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def occur(  # noqa: CFQ004
        self,
        sitem: 'HomieStreamItem',
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        :param sitem: Item containing information for operation.
        :returns: Boolean indicating the conditional outcomes.
        """

        plugin = self.plugin
        homie = plugin.homie
        childs = homie.childs
        scenes = childs.scenes
        groups = childs.groups

        params = self.params
        _scene = params.scene
        _group = params.group
        states = (
            params.states
            or _STATES)

        scene = scenes[_scene]
        group = groups[_group]


        model = PhueStreamItem

        if not isinstance(sitem, model):
            return NCFalse


        event = sitem.event

        type = event.get('type')

        if type != 'scene':
            return False


        origin = group.origin

        if origin is None:
            return False

        source = scene.source(
            origin, group)

        if source is None:
            return False

        service = event['id']


        intend = source['id']

        if service != intend:
            return False


        _state = getate(
            event, 'status/active')

        if _state is None:
            return NCFalse

        if _state != 'inactive':
            _state = 'active'


        return _state in states



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import TYPE_CHECKING
from typing import Type

from encommon.types import DictStrAny

from .aspire import HomieAspire
from .desire import HomieDesire
from .device import HomieDevice
from .group import HomieGroup
from .origin import HomieOrigin
from .scene import HomieScene
from ...utils import InvalidChild

if TYPE_CHECKING:
    from ..homie import Homie



HomieOrigins = dict[str, HomieOrigin]
HomieDevices = dict[str, HomieDevice]

HomieGroups = dict[str, HomieGroup]
HomieScenes = dict[str, HomieScene]

HomieAspires = dict[str, HomieAspire]
HomieDesires = dict[str, HomieDesire]



class HomieChilds:
    """
    Contain the object instances for related Homie children.

    :param homie: Primary class instance for Homie Automate.
    """

    __homie: 'Homie'

    __origins: HomieOrigins
    __devices: HomieDevices

    __groups: HomieGroups
    __scenes: HomieScenes

    __desires: HomieDesires
    __aspires: HomieAspires


    def __init__(
        self,
        homie: 'Homie',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie.logger.log_d(
            base=self,
            status='initial')

        self.__homie = homie

        self.__origins = {}
        self.__devices = {}

        self.__groups = {}
        self.__scenes = {}

        self.__desires = {}
        self.__aspires = {}

        self.build_objects()

        homie.logger.log_i(
            base=self,
            status='created')


    def build_objects(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        self.__build_origins()
        self.__build_devices()

        self.__build_groups()
        self.__build_scenes()

        self.__build_desires()
        self.__build_aspires()


    def __build_origins(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        homie = self.__homie
        params = homie.params
        origins = params.origins

        if origins is None:
            return None


        childs: HomieOrigins = {}


        items = origins.items()

        for name, origin in items:

            child = (
                self.
                get_origin(name))

            childs[name] = child(
                homie, name, origin)


        self.__origins = childs


    def __build_devices(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        homie = self.__homie
        params = homie.params
        devices = params.devices

        if devices is None:
            return None


        childs: HomieDevices = {}


        items = devices.items()

        for name, device in items:

            child = (
                self
                .get_device(name))

            childs[name] = child(
                homie, name, device)


        self.__devices = childs


    def __build_groups(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        homie = self.__homie
        params = homie.params
        groups = params.groups

        if groups is None:
            return None

        model = HomieGroup


        childs: HomieGroups = {}


        items = groups.items()

        for name, group in items:

            object = model(
                homie, name, group)

            childs[name] = object


        self.__groups = childs


    def __build_scenes(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        homie = self.__homie
        params = homie.params
        scenes = params.scenes

        if scenes is None:
            return None

        model = HomieScene


        childs: HomieScenes = {}


        items = scenes.items()

        for name, scene in items:

            object = model(
                homie, name, scene)

            childs[name] = object


        self.__scenes = childs


    def __build_desires(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        homie = self.__homie
        params = homie.params
        desires = params.desires

        if desires is None:
            return None

        model = HomieDesire


        childs: HomieDesires = {}


        items = desires.items()

        for name, desire in items:

            object = model(
                homie, name, desire)

            childs[name] = object


        self.__desires = childs


    def __build_aspires(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        homie = self.__homie
        params = homie.params
        aspires = params.aspires

        if aspires is None:
            return None

        model = HomieAspire


        childs: HomieAspires = {}


        items = aspires.items()

        for name, aspire in items:

            object = model(
                homie, name, aspire)

            childs[name] = object


        self.__aspires = childs


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """


        origins = (
            self.__origins
            .values())

        for origin in origins:
            origin.validate()


        devices = (
            self.__devices
            .values())

        for device in devices:
            device.validate()


        groups = (
            self.__groups
            .values())

        for group in groups:
            group.validate()


        scenes = (
            self.__scenes
            .values())

        for scene in scenes:
            scene.validate()


        desires = (
            self.__desires
            .values())

        for desire in desires:
            desire.validate()


        aspires = (
            self.__aspires
            .values())

        for aspire in aspires:
            aspire.validate()


    @property
    def origins(
        self,
    ) -> HomieOrigins:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        origins = self.__origins

        return dict(origins)


    @property
    def devices(
        self,
    ) -> HomieDevices:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        devices = self.__devices

        return dict(devices)


    @property
    def groups(
        self,
    ) -> HomieGroups:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        groups = self.__groups

        return dict(groups)


    @property
    def scenes(
        self,
    ) -> HomieScenes:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        scenes = self.__scenes

        return dict(scenes)


    @property
    def desires(
        self,
    ) -> HomieDesires:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        desires = self.__desires

        return dict(desires)


    @property
    def aspires(
        self,
    ) -> HomieAspires:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        aspires = self.__aspires

        return dict(aspires)


    @property
    def dumped(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        origins = self.origins
        devices = self.devices
        groups = self.groups
        scenes = self.scenes
        desires = self.desires
        aspires = self.aspires

        dumped: DictStrAny = {

            'origins': {
                k: v.dumped for k, v
                in origins.items()},

            'devices': {
                k: v.dumped for k, v
                in devices.items()},

            'groups': {
                k: v.dumped for k, v
                in groups.items()},

            'scenes': {
                k: v.dumped for k, v
                in scenes.items()},

            'desires': {
                k: v.dumped for k, v
                in desires.items()},

            'aspires': {
                k: v.dumped for k, v
                in aspires.items()}}

        return deepcopy(dumped)


    def get_origin(
        self,
        name: str,
    ) -> Type['HomieOrigin']:
        """
        Return the Homie class definition for its instantiation.

        :param name: Name of the object within the Homie config.
        :returns: Homie class definition for its instantiation.
        """

        from ...hubitat import HubiOrigin
        from ...philips import PhueOrigin
        from ...ubiquiti import UbiqOrigin


        homie = self.__homie
        params = homie.params
        origins = params.origins

        assert origins is not None

        origin = origins[name]


        _hubitat = origin.hubitat
        _philips = origin.philips
        _ubiquiti = origin.ubiquiti

        if _hubitat is not None:
            return HubiOrigin

        if _philips is not None:
            return PhueOrigin

        if _ubiquiti is not None:
            return UbiqOrigin


        raise InvalidChild(
            name, phase='initial')


    def get_device(
        self,
        name: str,
    ) -> Type['HomieDevice']:
        """
        Return the Homie class definition for its instantiation.

        :param name: Name of the object within the Homie config.
        :returns: Homie class definition for its instantiation.
        """

        from ...hubitat import HubiDevice
        from ...philips import PhueDevice
        from ...ubiquiti import UbiqDevice


        homie = self.__homie
        params = homie.params
        origins = self.__origins
        devices = params.devices

        assert devices is not None

        device = devices[name]
        origin = origins[
            device.origin]


        family = origin.family

        if family == 'hubitat':
            return HubiDevice

        if family == 'philips':
            return PhueDevice

        if family == 'ubiquiti':
            return UbiqDevice


        raise InvalidChild(
            name, phase='initial')



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..queue import HomieQueue
from ..queue import HomieQueueItem

if TYPE_CHECKING:
    from ...homie import Homie



def test_HomieQueueItem() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    item = HomieQueueItem()

    assert item.time.since < 1



def test_HomieQueue(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    _queue = HomieQueue[
        HomieQueueItem]

    item = HomieQueueItem()


    queue: _queue = (
        HomieQueue(homie))


    attrs = lattrs(queue)

    assert attrs == [
        '_HomieQueue__homie',
        '_HomieQueue__queue']


    assert inrepr(
        'queue.HomieQueue',
        queue)

    assert isinstance(
        hash(queue), int)

    assert instr(
        'queue.HomieQueue',
        queue)


    assert queue.empty

    assert queue.qsize == 0


    queue.put(item)

    assert not queue.empty

    assert queue.qsize == 1

    _item = queue.get()

    assert queue.empty

    assert queue.qsize == 0

    assert item == _item



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from ..action import arguments



def test_arguments() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    sargs = arguments([
        '--config', 'path',
        '--group', 'group',
        '--scene', 'scene'])

    assert sargs == {
        'config': 'path',
        'console': False,
        'debug': False,
        'dryrun': False,
        'potent': None,
        'paction': False,
        'group': 'group',
        'device': None,
        'state': None,
        'color': None,
        'level': None,
        'scene': 'scene'}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from encommon.types import BaseModel



class HomieParamsModel(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import lattrs

from ..device import HubiDevice

if TYPE_CHECKING:
    from ...homie import Homie
    from ...utils import TestBodies



def test_HubiDevice(
    homie: 'Homie',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    devices = childs.devices


    planets = bodies.planets

    family = 'hubitat'

    for planet in planets:

        name = f'{planet}_light2'

        device = devices[name]

        assert isinstance(
            device, HubiDevice)


        attrs = lattrs(device)

        assert attrs == [
            '_HomieChild__homie',
            '_HomieChild__name',
            '_HomieChild__params']


        device.validate()

        assert device.homie

        assert device.name == name

        assert device.family == family

        assert device.kind == 'device'

        assert device.origin

        assert device.params

        assert not device.source

        assert device.dumped


        device.origin.refresh()

        assert device.source



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from threading import Thread
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import clsname

from ..addons import HomieQueue
from ..addons import HomieQueueItem

if TYPE_CHECKING:
    from ..childs import HomieOrigin
    from ..homie import Homie
    from ..members import HomieMember
    from ..threads import HomieActionItem
    from ..threads import HomieStreamItem
    from ..threads import HomieUpdateItem



_UPDATES = dict[
    'HomieOrigin',
    'HomieUpdateItem']

_CONGEST = list[tuple[str, int]]



@dataclass
class HomieThreadItem(HomieQueueItem):
    """
    Contain information for sharing using the Python queue.
    """

    origin: str


    def __init__(
        self,
        origin: 'HomieOrigin',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.origin = origin.name

        super().__init__()



class HomieThread(Thread):
    """
    Common methods and routines for Homie Automate threads.

    :param member: Child class instance for Homie Automate.
    :param origin: Child class instance for Homie Automate.
    """

    __member: 'HomieMember'
    __origin: 'HomieOrigin'

    __aqueue: HomieQueue['HomieActionItem']
    __uqueue: HomieQueue['HomieUpdateItem']
    __squeue: HomieQueue['HomieStreamItem']


    def __init__(
        self,
        member: 'HomieMember',
        origin: 'HomieOrigin',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie = member.homie

        super().__init__()

        self.name = (
            f'{clsname(self)}'
            f'/{origin.name}')

        homie.logger.log_d(
            base=self,
            name=origin,
            status='initial')

        self.__member = member
        self.__origin = origin

        self.__build_objects()

        homie.logger.log_i(
            base=self,
            name=origin,
            status='created')


    def __build_objects(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        member = self.__member
        homie = member.homie

        self.__aqueue = (
            HomieQueue(homie))

        self.__uqueue = (
            HomieQueue(homie))

        self.__squeue = (
            HomieQueue(homie))


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.member.homie


    @property
    def member(
        self,
    ) -> 'HomieMember':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__member


    @property
    def origin(
        self,
    ) -> 'HomieOrigin':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__origin


    @property
    def aqueue(
        self,
    ) -> HomieQueue['HomieActionItem']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__aqueue


    @property
    def uqueue(
        self,
    ) -> HomieQueue['HomieUpdateItem']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__uqueue


    @property
    def squeue(
        self,
    ) -> HomieQueue['HomieStreamItem']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__squeue


    def expired(
        self,
        item: HomieThreadItem,
    ) -> bool:
        """
        Return the boolean indicating whether the item expired.

        :param item: Item containing information for operation.
        :returns: Boolean indicating whether the item expired.
        """

        member = self.member
        homie = member.homie

        since = item.time.since
        origin = item.origin

        if since < 60:
            return False

        homie.logger.log_w(
            base=self,
            item=item,
            origin=origin,
            status='expired',
            age=int(since))

        return True


    @property
    def congest(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        congest: _CONGEST = []

        name = self.name

        aqueue = self.aqueue
        uqueue = self.uqueue
        squeue = self.squeue


        if aqueue.qsize > 5:

            append = (
                f'{name}/aqueue',
                aqueue.qsize)

            congest.append(append)

        if uqueue.qsize > 5:

            append = (
                f'{name}/uqueue',
                uqueue.qsize)

            congest.append(append)

        if squeue.qsize > 5:

            append = (
                f'{name}/squeue',
                squeue.qsize)

            congest.append(append)


        return sorted(congest)


    @property
    def enqueue(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        enqueue: _CONGEST = []

        name = self.name

        aqueue = self.aqueue
        uqueue = self.uqueue
        squeue = self.squeue


        if aqueue.qsize > 0:

            append = (
                f'{name}/aqueue',
                aqueue.qsize)

            enqueue.append(append)

        if uqueue.qsize > 0:

            append = (
                f'{name}/uqueue',
                uqueue.qsize)

            enqueue.append(append)

        if squeue.qsize > 0:

            append = (
                f'{name}/squeue',
                squeue.qsize)

            enqueue.append(append)


        return sorted(enqueue)


    def run(
        self,
    ) -> None:
        """
        Perform whatever operation is associated with the class.
        """

        member = self.__member
        origin = self.__origin
        homie = member.homie
        vacate = member.vacate
        cancel = member.cancel

        homie.logger.log_i(
            base=self,
            name=origin,
            status='started')


        def _continue() -> bool:

            if cancel.is_set():
                return False

            enqueue = bool(
                self.enqueue)

            if vacate.is_set():
                return enqueue

            return True


        while _continue():

            self.operate_updates()

            self.operate()

            block_sleep(0.1)


        homie.logger.log_i(
            base=self,
            name=origin,
            status='stopped')


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service threads.
        """

        raise NotImplementedError


    def operate_updates(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service threads.
        """

        member = self.__member
        homie = member.homie
        childs = homie.childs

        vacate = member.vacate
        uqueue = self.__uqueue

        origins = (
            childs.origins
            .values())


        updates: _UPDATES = {}


        def _set_update() -> None:

            name = uitem.origin
            _name = origin.name

            if name != _name:
                return None

            updates[origin] = uitem


        while not uqueue.empty:

            uitem = uqueue.get()

            if vacate.is_set():
                continue

            if self.expired(uitem):
                continue

            for origin in origins:
                _set_update()


        items = updates.items()

        for origin, uitem in items:

            (origin
             .set_update(uitem))


    def stop(
        self,
    ) -> None:
        """
        Wait for the thread object to complete routine and exit.
        """

        member = self.__member
        origin = self.__origin
        homie = member.homie

        homie.logger.log_d(
            base=self,
            name=origin,
            status='waiting')

        self.join()

        homie.logger.log_d(
            base=self,
            name=origin,
            status='awaited')



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .button import DriverPhueButtonParams
from .change import DriverPhueChangeParams
from .contact import DriverPhueContactParams
from .motion import DriverPhueMotionParams
from .scene import DriverPhueSceneParams



__all__ = [
    'DriverPhueButtonParams',
    'DriverPhueContactParams',
    'DriverPhueMotionParams',
    'DriverPhueChangeParams',
    'DriverPhueSceneParams']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import DictStrAny
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils.sample import ENPYRWS

from . import SAMPLES
from ..config import HomieConfig
from ..homie import Homie
from ..threads import HomieThreadItem

if TYPE_CHECKING:
    from ...utils import TestBodies



def test_Homie(
    homie: Homie,
    replaces: DictStrAny,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param replaces: Mapping of what to replace in samples.
    """


    attrs = lattrs(homie)

    assert attrs == [
        '_Homie__config',
        '_Homie__logger',
        '_Homie__persist',
        '_Homie__childs',
        '_Homie__desired',
        '_Homie__aspired']


    assert inrepr(
        'homie.Homie',
        homie)

    assert isinstance(
        hash(homie), int)

    assert instr(
        'homie.Homie',
        homie)


    assert homie.config

    assert homie.logger

    assert homie.persist

    assert homie.childs

    assert homie.params

    assert not homie.dryrun

    assert homie.potent

    assert homie.desired

    assert homie.aspired


    assert homie.refresh()


    sample_path = (
        SAMPLES / 'dumped.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=homie.dumped,
        replace=replaces)

    expect = prep_sample(
        content=homie.dumped,
        replace=replaces)

    assert expect == sample



def test_Homie_printer(
    homie: Homie,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    childs = homie.childs
    origins = childs.origins

    origin = origins[
        'jupiter_philips']

    model = HomieThreadItem

    item = model(origin)

    homie.printer(item)



def test_Homie_actions(
    homie: Homie,
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param bodies: Locations and groups for use in testing.
    """

    childs = homie.childs
    devices = childs.devices
    groups = childs.groups


    assert homie.refresh()


    moons = (
        dict(bodies.moons)
        .items())

    for planet, moon in moons:


        device = devices[
            f'{planet}_light2']

        aitems = (
            homie.get_actions(
                target=device,
                state='poweron',
                color='ff00cc',
                level=69))

        assert len(aitems) == 1

        assert homie.set_actions(
            aitems, force=True)

        assert homie.set_actions(
            aitems, change=False)


        device = devices[
            f'{planet}_light1']

        aitems = (
            homie.get_actions(
                target=device,
                state='nopower',
                color='ff00cc',
                level=100))

        assert len(aitems) == 1

        assert homie.set_actions(
            aitems, force=True)

        assert homie.set_actions(
            aitems, change=False)


        group = groups[moon]

        aitems = (
            homie.get_actions(
                target=group,
                state='poweron',
                color='ff00cc',
                level=69))

        assert len(aitems) == 2

        assert homie.set_actions(
            aitems, force=True)

        assert homie.set_actions(
            aitems, change=False)


        aitems = (
            homie
            .get_actions(group))

        assert len(aitems) == 0



def test_Homie_cover(
) -> None:
    """
    Perform various tests associated with relevant routines.
    """

    Homie(HomieConfig())



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Optional

from pydantic import Field

from .plugin import HomiePluginParams
from ...builtins.params import DriverBltnPeriodParams
from ...builtins.params import DriverBltnStoreParams
from ...philips.params import DriverPhueChangeParams
from ...ubiquiti.params import DriverUbiqClientParams



class HomieWhereParams(HomiePluginParams, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    .. note::
       When using the `default` `family`, the conditons are
       in an `AND` relationship. If another name is given
       for `family`, those will be with an OR relationship.
    """

    negate: Annotated[
        bool,
        Field(False,
              description='Invert conditional outcome')]

    family: Annotated[
        str,
        Field('default',
              description='Combine conditions in group',
              min_length=1)]

    builtins_store: Annotated[
        Optional[DriverBltnStoreParams],
        Field(None,
              description='Plugin for the operations')]

    builtins_period: Annotated[
        Optional[DriverBltnPeriodParams],
        Field(None,
              description='Plugin for the operations')]

    philips_change: Annotated[
        Optional[DriverPhueChangeParams],
        Field(None,
              description='Plugin for the operations')]

    ubiquiti_client: Annotated[
        Optional[DriverUbiqClientParams],
        Field(None,
              description='Plugin for the operations')]



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Optional

from pydantic import Field

from ...plugins.scene import PhueSceneState
from ....homie.params.common import HomieParamsModel



class DriverPhueSceneParams(HomieParamsModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    scene: Annotated[
        str,
        Field(...,
              description='Scene that is in scope',
              min_length=1)]

    group: Annotated[
        str,
        Field(...,
              description='Group that is in scope',
              min_length=1)]

    states: Annotated[
        Optional[list[PhueSceneState]],
        Field(None,
              description='States that will be matched',
              min_length=1)]


    def __init__(
        # NOCVR
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        states = data.get('states')

        if isinstance(states, str):
            data['states'] = [states]

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Time

if TYPE_CHECKING:
    from ..childs import HomieDesire
    from ..childs import HomieScene
    from ..common import HomieState
    from ..homie import Homie
    from ..threads import HomieActionNode



@dataclass
class HomieDesiredItem:
    """
    Contain information related to the desired child state.
    """

    target: 'HomieActionNode'
    weight: int

    desire: 'HomieDesire'

    state: Optional['HomieState'] = None
    color: Optional[str] = None
    level: Optional[int] = None
    scene: Optional['HomieScene'] = None


    def __init__(
        self,
        target: 'HomieActionNode',
        desire: 'HomieDesire',
    ) -> None:
        """
        Initialize instance for class using provided parameters.

        .. note::
           Somewhat similar to same method in HomieAspiredItem.
        """

        homie = desire.homie
        childs = homie.childs

        params = desire.params
        weight = params.weight

        self.target = target
        self.weight = weight

        self.desire = desire

        scene = params.scene
        stage = params.stage

        self.state = None
        self.color = None
        self.level = None
        self.scene = None


        if scene is not None:

            _scene = (
                childs.scenes
                [scene])

            self.scene = _scene


        if stage is not None:

            _state = stage.state
            _color = stage.color
            _level = stage.level

            self.state = _state
            self.color = _color
            self.level = _level


    def __lt__(
        self,
        other: 'HomieDesiredItem',
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        .. note::
           Useful with sorting to influence consistent output.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        name = self.target.name
        _name = other.target.name

        return bool(name < _name)


    @property
    def paused(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether operation delayed.

        :returns: Boolean indicating whether operation delayed.
        """

        target = self.target
        desire = self.desire

        return desire.paused(target)


    @property
    def applied(
        self,
    ) -> bool:
        """
        Return the boolean indicating that there will be change.

        :returns: Boolean indicating that there will be change.
        """

        desire = self.desire
        homie = desire.homie
        potent = homie.potent

        target = self.target
        state = self.state
        color = self.color
        level = self.level
        scene = self.scene

        aitems = (
            homie.get_actions(
                target=target,
                state=state,
                color=color,
                level=level,
                scene=scene))

        return not (
            homie.set_actions(
                aitems=aitems,
                force=potent,
                change=False))


    def matched(
        self,
    ) -> None:
        """
        Perform operations when this desire is the most desired.
        """

        target = self.target
        desire = self.desire

        return desire.matched(target)


    def omitted(
        self,
    ) -> None:
        """
        Perform operations when this desire is the most desired.
        """

        target = self.target
        desire = self.desire

        return desire.omitted(target)



_DESIRED = list[HomieDesiredItem]
_STAGING = dict['HomieActionNode', set['HomieDesire']]
_MATCHED = set['HomieDesire']
_TARGETS = dict['HomieActionNode', 'HomieDesiredItem']



class HomieDesired:
    """
    Contain information related to the desired child state.

    :param homie: Primary class instance for Homie Automate.
    """

    __homie: 'Homie'


    def __init__(
        self,
        homie: 'Homie',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__homie = homie


    def items(  # noqa: CFQ001
        self,
        time: Time,
    ) -> _DESIRED:
        """
        Return the action state objects for the related targets.

        ..note::
          Somewhat similar to same method within HomieAspire.

        :param time: Time that will be used in the conditionals.
        :returns: Action state objects for the related targets.
        """

        homie = self.__homie
        persist = homie.persist
        childs = homie.childs

        desires = (
            childs.desires
            .values())

        desired: _DESIRED = []
        prepare: _STAGING = {}
        matched: _MATCHED = set()
        targets: _TARGETS = {}

        model = HomieDesiredItem


        for desire in desires:

            when = desire.when(time)

            if when is False:
                desire.omitted()
                continue

            matched.add(desire)


            devices = desire.devices

            for device in devices:

                if not prepare.get(device):
                    prepare[device] = set()

                (prepare[device]
                 .add(desire))


            groups = desire.groups

            for group in groups:

                if not prepare.get(group):
                    prepare[group] = set()

                (prepare[group]
                 .add(desire))


        items = prepare.items()

        for target, serised in items:

            _serised = sorted(
                serised,
                key=lambda x: x.weight,
                reverse=True)


            for desire in _serised[1:]:

                object = model(
                    target, desire)

                object.omitted()


            desire = _serised[0]

            object = model(
                target, desire)


            if object.applied:

                object.omitted()

                homie.logger.log_d(
                    base=self,
                    item=desire,
                    name=desire,
                    target=target.name,
                    status='omit')

                continue

            if object.paused:

                homie.logger.log_d(
                    base=self,
                    item=desire,
                    name=desire,
                    target=target.name,
                    status='pause')

                continue

            homie.logger.log_d(
                base=self,
                item=desire,
                name=desire,
                target=target.name,
                status='match')

            object.matched()

            targets[target] = object


        for desire in matched:

            store = (
                desire.params
                .store)

            if store is None:
                continue

            if homie.dryrun:
                continue

            for value in store:

                persist.insert(
                    value.unique,
                    value.value,
                    value.expire)


        objects = targets.values()

        for object in objects:

            desired.append(object)


        return sorted(desired)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from threading import current_thread
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from encommon.colors import Color
from encommon.times import Time
from encommon.types import DictStrAny

from enconnect.hubitat import Bridge

from .helpers import HubiFetch
from .helpers import HubiMerge
from .helpers import merge_fetch
from .helpers import merge_find
from .helpers import request_action
from .models import HubiModels
from ..homie.childs import HomieOrigin
from ..utils import Idempotent
from ..utils import MultipleSource

if TYPE_CHECKING:
    from ..homie.childs import HomieScene
    from ..homie.common import HomieKinds
    from ..homie.common import HomieState
    from ..homie.threads import HomieActionNode
    from ..homie.threads import HomieActionItem
    from ..homie.threads import HomieUpdateBase
    from ..homie.threads import HomieUpdateItem



class HubiOrigin(HomieOrigin):
    """
    Contain the relevant attributes from the related source.
    """

    __bridge: Bridge

    __fetch: Optional[HubiFetch]
    __merge: Optional[HubiMerge]


    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(
            *args, **kwargs)


        params = (
            self.params
            .hubitat)

        assert params is not None

        bridge = Bridge(
            params.bridge)


        self.__bridge = bridge

        self.__fetch = None
        self.__merge = None


    @property
    def family(
        self,
    ) -> Literal['hubitat']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'hubitat'


    @property
    def bridge(
        self,
    ) -> Bridge:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__bridge


    @property
    def fetch(
        self,
    ) -> Optional[HubiFetch]:
        """
        Return the content related to the item in parent system.

        :returns: Content related to the item in parent system.
        """

        fetch = self.__fetch

        return deepcopy(fetch)


    @property
    def merge(
        self,
    ) -> Optional[HubiMerge]:
        """
        Return the content related to the item in parent system.

        :returns: Content related to the item in parent system.
        """

        fetch = self.__fetch
        merge = self.__merge

        if merge is not None:
            return deepcopy(merge)

        if fetch is None:
            return None

        merge = (
            merge_fetch(fetch))

        self.__merge = merge

        return deepcopy(merge)


    def refresh(
        self,
        timeout: Optional[int] = None,
    ) -> bool:
        """
        Refresh the cached information for the remote upstream.

        :param timeout: Timeout waiting for the server response.
        :returns: Boolean indicating the success of the refresh.
        """

        homie = self.homie
        bridge = self.__bridge
        request = bridge.request

        runtime = Time()

        try:

            response = request(
                'get', 'devices/all',
                timeout=timeout)

            (response
             .raise_for_status())


            fetch = response.json()

            assert isinstance(
                fetch, list)


            self.__fetch = fetch
            self.__merge = None


            homie.logger.log_d(
                base=self,
                name=self,
                item='refresh',
                elapsed=runtime,
                status='success')

            return True


        except Exception as reason:

            homie.logger.log_e(
                base=self,
                name=self,
                item='refresh',
                elapsed=runtime,
                status='exception',
                exc_info=reason)

            return False


    def get_action(
        self,
        target: 'HomieActionNode',
        *,
        state: Optional['HomieState'] = None,
        color: Optional[str | Color] = None,
        level: Optional[int] = None,
        scene: Optional[Union[str, 'HomieScene']] = None,
    ) -> 'HomieActionItem':
        """
        Return the new item containing information for operation.

        .. note::
           This method is identical to same with other origins.

        :param target: Device or group settings will be updated.
        :param state: Determine the state related to the target.
        :param color: Determine the color related to the target.
        :param level: Determine the level related to the target.
        :param scene: Determine the scene related to the target.
        :returns: New item containing information for operation.
        """

        homie = self.homie
        childs = homie.childs
        scenes = childs.scenes

        model = (
            HubiModels
            .action())

        if isinstance(color, str):
            color = Color(color)

        if isinstance(scene, str):
            scene = scenes[scene]

        return model(
            self, target,
            state=state,
            color=color,
            level=level,
            scene=scene)


    def set_action(  # noqa: CFQ004
        self,
        target: 'HomieActionNode',
        aitem: 'HomieActionItem',
        force: bool = False,
        change: bool = True,
        timeout: Optional[int] = None,
    ) -> bool:
        """
        Perform the provided action with specified Homie target.

        :param target: Device or group settings will be updated.
        :param aitem: Item containing information for operation.
        :param force: Override the default for full idempotency.
        :param change: Determine whether the change is executed.
        :param timeout: Timeout waiting for the server response.
        :returns: Boolean indicating whether or not was changed.
        """

        homie = self.homie
        childs = homie.childs
        scenes = childs.scenes

        runtime = Time()

        thread = current_thread()


        name = aitem.scene

        scene = (
            scenes[name]
            if name is not None
            else None)


        try:

            request_action(
                self, target,
                state=aitem.state,
                color=aitem.color,
                level=aitem.level,
                scene=scene,
                force=force,
                change=change,
                timeout=timeout)


        except Idempotent:

            if change is False:
                return False

            homie.logger.log_d(
                base=self,
                name=self,
                item='action',
                target=target.name,
                thread=thread.name,
                elapsed=runtime,
                status='nochange')

            return False


        except Exception as reason:

            homie.logger.log_e(
                base=self,
                name=self,
                item='action',
                target=target.name,
                thread=thread.name,
                elapsed=runtime,
                status='exception',
                exc_info=reason)

            return False

        else:

            if change is False:
                return True

            homie.logger.log_d(
                base=self,
                name=self,
                item='action',
                target=target.name,
                thread=thread.name,
                elapsed=runtime,
                status='success')

            return True


    def get_update(
        self,
    ) -> 'HomieUpdateItem':
        """
        Return the new item containing information for operation.

        :returns: New item containing information for operation.
        """

        model = (
            HubiModels
            .update())

        fetch = self.fetch

        assert fetch is not None

        return model(
            self, fetch)


    def set_update(
        self,
        uitem: 'HomieUpdateBase',
    ) -> bool:
        """
        Replace or update internal configuration with provided.

        :param uitem: Item containing information for operation.
        :returns: Boolean indicating whether or not was changed.
        """

        _name = uitem.origin

        assert _name == self.name


        assert hasattr(
            uitem, 'fetch')


        fetch = deepcopy(
            uitem.fetch)

        self.__fetch = fetch
        self.__merge = None

        return True


    def source(
        self,
        kind: Optional['HomieKinds'] = None,
        unique: Optional[str] = None,
        label: Optional[str] = None,
        relate: Optional['HomieActionNode'] = None,
    ) -> Optional[DictStrAny]:
        """
        Return the content related to the item in parent system.

        .. note::
           This method is redundant with other origin classes.

        :param kind: Which kind of Homie object will be located.
        :param unique: Unique identifier within parents system.
        :param label: Friendly name or label within the parent.
        :param relate: Child class instance for Homie Automate.
        :returns: Content related to the item in parent system.
        """

        merge = self.__merge

        if merge is None:
            merge = self.merge

        if merge is None:
            return None

        assert relate is None

        found = merge_find(
            merge=merge,
            kind=kind,
            unique=unique,
            label=label)

        if len(found) == 0:
            return None

        if len(found) == 1:
            return deepcopy(found[0])

        raise MultipleSource



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Callable
from typing import Optional

from enconnect.philips import BridgeParams

from pydantic import Field

from ...homie.params.common import HomieParamsModel



class PhueOriginParams(HomieParamsModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    bridge: Annotated[
        BridgeParams,
        Field(...,
              description='Connection specific parameters')]


    def __init__(
        # NOCVR
        self,
        /,
        _parse: Optional[Callable[..., Any]] = None,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


        if _parse is not None:

            data['bridge'] = _parse(
                data['bridge'])


        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.times import Time

if TYPE_CHECKING:
    from ...service import HomieService
    from ....utils import TestBodies



def test_HomieThread_cover(
    service: 'HomieService',
    bodies: 'TestBodies',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Homie Automate class instance.
    :param bodies: Locations and groups for use in testing.
    """

    homie = service.homie
    childs = homie.childs
    origins = childs.origins

    member = service.actions
    threads = member.threads


    planets = bodies.planets

    for planet in planets:

        origin = origins[
            f'{planet}_philips']

        thread = threads[
            f'{planet}_philips']


        assert origin.refresh()

        uitem = origin.get_update()

        expired = (
            thread.expired(uitem))

        assert expired is False

        uitem.time = Time('-1h')

        expired = (
            thread.expired(uitem))

        assert expired is True



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING
from typing import get_args

from encommon.types import NCFalse
from encommon.types import getate

from .helpers import phue_sensors
from ..models import PhueModels
from ..stream import PhueStreamItem
from ...homie.plugins import HomieDriver
from ...utils import InvalidParam

if TYPE_CHECKING:
    from ..params import DriverPhueButtonParams
    from ...homie.plugins import HomiePluginKinds
    from ...homie.threads import HomieStreamItem



PhueButtonEvent = Literal[
    'initial_press',
    'long_release',
    'short_release']

PhueButtonSensor = Literal[
    'button1',
    'button2',
    'button3',
    'button4']

_EVENTS = list(
    get_args(PhueButtonEvent))

_SENSORS = list(
    get_args(PhueButtonSensor))



class DriverPhueButton(HomieDriver):
    """
    Match specific conditions for determining desired state.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        plugin = self.plugin
        family = self.family
        homie = plugin.homie
        childs = homie.childs
        devices = childs.devices
        params = self.params

        _device = params.device


        if _device not in devices:

            raise InvalidParam(
                param='device',
                value=_device,
                error='noexist')


        device = devices[_device]

        _family = device.family

        if _family != family:

            raise InvalidParam(
                param='device',
                value=_device,
                error='invalid')


    @property
    def family(
        self,
    ) -> Literal['philips']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'philips'


    @property
    def kinds(
        self,
    ) -> list['HomiePluginKinds']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return ['occur']


    @property
    def params(
        self,
    ) -> 'DriverPhueButtonParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            PhueModels
            .drivers()
            .button())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def occur(  # noqa: CFQ004
        self,
        sitem: 'HomieStreamItem',
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        :param sitem: Item containing information for operation.
        :returns: Boolean indicating the conditional outcomes.
        """

        plugin = self.plugin
        homie = plugin.homie
        childs = homie.childs
        devices = childs.devices

        params = self.params
        target = params.device
        sensor = params.sensor
        events = (
            params.events
            or _EVENTS)


        model = PhueStreamItem

        if not isinstance(sitem, model):
            return NCFalse


        event = sitem.event

        if 'button' not in event:
            return False


        source = (
            devices[target]
            .source)

        if source is None:
            return False

        service = event['id']


        sensors = (
            phue_sensors(source))

        if sensor not in sensors:
            return NCFalse

        intend = sensors[sensor]

        if service != intend:
            return False


        _event = getate(
            event,
            'button/last_event')

        if _event not in events:
            return False


        return True



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING
from typing import Type

from .plugin import HomiePlugin
from ..models import HomieModels
from ...utils import InvalidParam
from ...utils import UnexpectedCondition

if TYPE_CHECKING:
    from .driver import HomieDriver
    from ..params import HomieOccurParams
    from ..threads import HomieStreamItem



class HomieOccur(HomiePlugin):
    """
    Match specific conditions for determining desired state.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        params = self.params

        _bltn_regexp = params.builtins_regexp
        _phue_button = params.philips_button
        _phue_contact = params.philips_contact
        _phue_motion = params.philips_motion
        _phue_scene = params.philips_scene

        super().validate()


        enabled = [
            1 if _bltn_regexp else 0,
            1 if _phue_button else 0,
            1 if _phue_contact else 0,
            1 if _phue_motion else 0,
            1 if _phue_scene else 0]

        if sum(enabled) != 1:

            raise InvalidParam(
                error='minimal')


    @property
    def kind(
        self,
    ) -> Literal['occur']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'occur'


    @property
    def params(
        self,
    ) -> 'HomieOccurParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            HomieModels
            .occur())

        params = super().params

        assert isinstance(
            params, model)

        return params


    def get_drivers(
        self,
    ) -> dict[str, Type['HomieDriver']]:
        """
        Return the Homie class definition for its instantiation.

        :returns: Homie class definition for its instantiation.
        """

        from ...builtins import (
            DriverBltnRegexp)

        from ...philips import (
            DriverPhueButton)

        from ...philips import (
            DriverPhueContact)

        from ...philips import (
            DriverPhueMotion)

        from ...philips import (
            DriverPhueScene)


        return {

            'builtins_regexp': (
                DriverBltnRegexp),

            'philips_button': (
                DriverPhueButton),

            'philips_contact': (
                DriverPhueContact),

            'philips_motion': (
                DriverPhueMotion),

            'philips_scene': (
                DriverPhueScene)}


    def when(
        self,
        sitem: 'HomieStreamItem',
    ) -> bool:
        """
        Return the boolean indicating the conditional outcomes.

        .. note::
           Somewhat similar to same method within HomieWhere.

        :param sitem: Item containing information for operation.
        :returns: Boolean indicating the conditional outcomes.
        """

        drivers = self.drivers

        occurs: list[bool] = []


        for driver in drivers:

            occur = driver.occur(sitem)

            occurs.append(occur)


        if len(occurs) >= 1:
            return any(occurs)

        raise UnexpectedCondition



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any

from pydantic import Field

from ...homie.params.common import HomieParamsModel



class DriverBltnRegexpParams(HomieParamsModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    patterns: Annotated[
        list[str],
        Field(...,
              description='Regular expression match',
              min_length=1)]

    complete: Annotated[
        bool,
        Field(False,
              description='Perform complete match')]


    def __init__(
        # NOCVR
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        patterns = data.get('patterns')

        if isinstance(patterns, str):
            data['patterns'] = [patterns]

        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Optional

from pydantic import Field

from .child import HomieChildParams



class HomieDeviceParams(HomieChildParams, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.
    """

    origin: Annotated[
        str,
        Field(...,
              description='Origin where device exists',
              min_length=1)]

    label: Annotated[
        Optional[str],
        Field(None,
              description='Device name in the origin',
              min_length=1)]

    unique: Annotated[
        Optional[str],
        Field(None,
              description='Unique identifier in origin',
              min_length=1)]
